# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from typing import Callable
from unittest.mock import ANY, MagicMock, call, patch

import numpy as np
import pandapower as pp
import pandas as pd
import pytest
from power_grid_model import Branch3Side, BranchSide, LoadGenType, WindingType, initialize_array

from power_grid_model_io.converters.pandapower_converter import PandaPowerConverter

from ...utils import MockDf, MockFn, assert_struct_array_equal


def _generate_ids(*args, **kwargs):
    return MockFn("_generate_ids", *args, **kwargs)


def _get_pgm_ids(*args, **kwargs):
    return MockFn("_get_pgm_ids", *args, **kwargs)


def _get_pp_attr(*args, **kwargs):
    return MockFn("_get_pp_attr", *args, **kwargs)


def get_switch_states(*args, **kwargs):
    return MockFn("get_switch_states", *args, **kwargs)


def get_trafo3w_switch_states(*args, **kwargs):
    return MockFn("get_trafo3w_switch_states", *args, **kwargs)


def get_trafo_winding_types(*args, **kwargs):
    return MockFn("get_trafo_winding_types", *args, **kwargs)


def get_trafo3w_winding_types(*args, **kwargs):
    return MockFn("get_trafo3w_winding_types", *args, **kwargs)


def _get_transformer_tap_side(*args, **kwargs):
    return MockFn("_get_transformer_tap_side", *args, **kwargs)


def _get_3wtransformer_tap_side(*args, **kwargs):
    return MockFn("_get_3wtransformer_tap_side", *args, **kwargs)


def _get_3wtransformer_tap_size(*args, **kwargs):
    return MockFn("_get_3wtransformer_tap_size", *args, **kwargs)


def _get_tap_size(*args, **kwargs):
    return MockFn("_get_tap_size", *args, **kwargs)


def np_array(*args, **kwargs):
    return MockFn("np.array", *args, **kwargs)


@pytest.fixture
def converter() -> PandaPowerConverter:
    converter = PandaPowerConverter()
    converter._generate_ids = MagicMock(side_effect=_generate_ids)  # type: ignore
    converter._get_pgm_ids = MagicMock(side_effect=_get_pgm_ids)  # type: ignore
    converter._get_pp_attr = MagicMock(side_effect=_get_pp_attr)  # type: ignore
    converter.get_switch_states = MagicMock(side_effect=get_switch_states)  # type: ignore
    converter.get_trafo_winding_types = MagicMock(side_effect=get_trafo_winding_types)  # type: ignore
    converter._get_tap_size = MagicMock(side_effect=_get_tap_size)  # type: ignore
    converter.get_trafo_winding_types = MagicMock(side_effect=get_trafo_winding_types)  # type: ignore # TODO check this
    converter.get_trafo3w_switch_states = MagicMock(side_effect=get_trafo3w_switch_states)  # type: ignore
    converter.get_trafo3w_winding_types = MagicMock(side_effect=get_trafo3w_winding_types)  # type: ignore
    converter._get_transformer_tap_side = MagicMock(side_effect=_get_transformer_tap_side)  # type: ignore
    converter._get_3wtransformer_tap_side = MagicMock(side_effect=_get_3wtransformer_tap_side)  # type: ignore
    converter._get_3wtransformer_tap_size = MagicMock(side_effect=_get_3wtransformer_tap_size)  # type: ignore

    return converter


@pytest.fixture
def two_pp_objs() -> MockDf:
    return MockDf(2)


@patch("power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._fill_pgm_extra_info")
@patch("power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._fill_pp_extra_info")
@patch("power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._create_input_data")
def test_parse_data(
    create_input_data_mock: MagicMock, fill_pp_extra_info_mock: MagicMock, fill_pgm_extra_info_mock: MagicMock
):
    # Arrange
    converter = PandaPowerConverter()

    def create_input_data():
        converter.pgm_input_data = {"node": np.array([])}

    create_input_data_mock.side_effect = create_input_data

    # Act
    result = converter._parse_data(data={"bus": pd.DataFrame()}, data_type="input", extra_info=None)

    # Assert
    create_input_data_mock.assert_called_once_with()
    fill_pgm_extra_info_mock.assert_not_called()
    fill_pp_extra_info_mock.assert_not_called()
    assert len(converter.pp_input_data) == 1 and "bus" in converter.pp_input_data
    assert len(converter.pgm_input_data) == 1 and "node" in converter.pgm_input_data
    assert len(result) == 1 and "node" in result


@patch("power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._fill_pgm_extra_info")
@patch("power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._fill_pp_extra_info")
@patch("power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._create_input_data")
def test_parse_data__extra_info(
    create_input_data_mock: MagicMock, fill_pp_extra_info_mock: MagicMock, fill_pgm_extra_info_mock: MagicMock
):
    # Arrange
    converter = PandaPowerConverter()

    extra_info = MagicMock("extra_info")

    # Act
    converter._parse_data(data={}, data_type="input", extra_info=extra_info)

    # Assert
    create_input_data_mock.assert_called_once_with()
    fill_pgm_extra_info_mock.assert_called_once_with(extra_info=extra_info)
    fill_pp_extra_info_mock.assert_called_once_with(extra_info=extra_info)


def test_parse_data__update_data():
    # Arrange
    converter = PandaPowerConverter()

    # Act/Assert
    with pytest.raises(ValueError):
        converter._parse_data(data={}, data_type="update", extra_info=None)


def test_fill_pgm_extra_info():
    # Arrange
    converter = PandaPowerConverter()
    converter.idx_lookup[("bus", None)] = pd.Series([101, 102, 103], index=[0, 1, 2])
    converter.idx_lookup[("load", "const_current")] = pd.Series([201, 202, 203], index=[3, 4, 5])
    converter.pgm_input_data["sym_load"] = initialize_array("input", "sym_load", 3)
    converter.pgm_input_data["sym_load"]["id"] = [3, 4, 5]
    converter.pgm_input_data["sym_load"]["node"] = [0, 1, 2]
    converter.pgm_input_data["line"] = initialize_array("input", "line", 2)
    converter.pgm_input_data["line"]["id"] = [6, 7]
    converter.pgm_input_data["line"]["from_node"] = [0, 1]
    converter.pgm_input_data["line"]["to_node"] = [1, 2]
    converter.pgm_input_data["line"]["i_n"] = [106.0, 105.0]

    # Act
    extra_info = {}
    converter._fill_pgm_extra_info(extra_info=extra_info)

    # Assert
    assert len(extra_info) == 8
    assert extra_info[0] == {"id_reference": {"table": "bus", "index": 101}}
    assert extra_info[1] == {"id_reference": {"table": "bus", "index": 102}}
    assert extra_info[2] == {"id_reference": {"table": "bus", "index": 103}}
    assert extra_info[3] == {
        "id_reference": {"table": "load", "name": "const_current", "index": 201},
        "pgm_input": {"node": 0},
    }
    assert extra_info[4] == {
        "id_reference": {"table": "load", "name": "const_current", "index": 202},
        "pgm_input": {"node": 1},
    }
    assert extra_info[5] == {
        "id_reference": {"table": "load", "name": "const_current", "index": 203},
        "pgm_input": {"node": 2},
    }
    assert extra_info[6] == {"pgm_input": {"from_node": 0, "to_node": 1, "i_n": 106.0}}
    assert extra_info[7] == {"pgm_input": {"from_node": 1, "to_node": 2, "i_n": 105.0}}


def test_fill_pp_extra_info():
    # Arrange
    converter = PandaPowerConverter()
    converter.idx_lookup[("line", None)] = pd.Series([102, 103], index=[1, 2])
    converter.idx_lookup[("trafo", None)] = pd.Series([201, 202, 203], index=[3, 4, 5])
    converter.idx[("line", None)] = pd.Series([1, 2], index=[102, 103])
    converter.idx[("trafo", None)] = pd.Series([3, 4, 5], index=[201, 202, 203])

    converter.pp_input_data["trafo"] = pd.DataFrame(
        {"df": [0.1, 0.2, 0.3], "other": [0.1, 0.2, 0.3]}, index=[201, 202, 203]
    )
    converter.pp_input_data["line"] = pd.DataFrame([10, 11, 12], columns=["df"], index=[201, 202, 203])

    # Act
    extra_info = {}
    converter._fill_pp_extra_info(extra_info=extra_info)

    # Assert
    assert len(extra_info) == 3
    assert extra_info[3] == {"pp_input": {"df": 0.1}}
    assert extra_info[4] == {"pp_input": {"df": 0.2}}
    assert extra_info[5] == {"pp_input": {"df": 0.3}}


def test_fill_pp_extra_info__no_info():
    # Arrange
    converter = PandaPowerConverter()
    converter.idx_lookup[("trafo", None)] = pd.Series([201, 202, 203], index=[3, 4, 5])
    converter.idx[("trafo", None)] = pd.Series([3, 4, 5], index=[201, 202, 203])

    converter.pp_input_data["trafo"] = pd.DataFrame(
        {"col1": [0.1, 0.2, 0.3], "col2": [0.1, 0.2, 0.3]}, index=[201, 202, 203]
    )
    converter.pgm_input_data["transformer"] = initialize_array("input", "transformer", 3)
    converter.pgm_input_data["transformer"]["id"] = [3, 4, 5]
    # Act
    extra_info = {}
    converter._fill_pp_extra_info(extra_info=extra_info)

    # Assert
    assert extra_info == {}


@patch("power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._extra_info_to_idx_lookup")
@patch("power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._extra_info_to_pgm_input_data")
@patch("power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._create_output_data")
def test__serialize_data__sym(
    create_output_data_mock: MagicMock, extra_info_pgm_input_data: MagicMock, extra_info_to_idx_lookup: MagicMock
):
    # Arrange
    converter = PandaPowerConverter()
    line_sym_output_array = initialize_array("sym_output", "line", 1)

    def create_output_data():
        converter.pp_output_data = {"res_line": pd.DataFrame(np.array([]))}

    create_output_data_mock.side_effect = create_output_data

    # Act
    result = converter._serialize_data(data={"line": line_sym_output_array}, extra_info=None)

    # Assert
    create_output_data_mock.assert_called_once_with()
    extra_info_to_idx_lookup.assert_not_called()
    extra_info_pgm_input_data.assert_not_called()
    assert len(converter.pp_output_data) == 1 and "res_line" in converter.pp_output_data
    assert len(converter.pgm_output_data) == 1 and "line" in converter.pgm_output_data
    assert len(result) == 1 and "res_line" in result


@patch("power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._extra_info_to_idx_lookup")
@patch("power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._extra_info_to_pgm_input_data")
@patch("power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._create_output_data_3ph")
def test__serialize_data__asym(
    create_output_data_3ph_mock: MagicMock, extra_info_pgm_input_data: MagicMock, extra_info_to_idx_lookup: MagicMock
):
    # Arrange
    converter = PandaPowerConverter()
    line_asym_output_array = initialize_array("asym_output", "line", 1)

    def create_output_data_3ph():
        converter.pp_output_data = {"res_line_3ph": pd.DataFrame(np.array([]))}

    create_output_data_3ph_mock.side_effect = create_output_data_3ph

    # Act
    result = converter._serialize_data(data={"line": line_asym_output_array}, extra_info=None)

    # Assert
    create_output_data_3ph_mock.assert_called_once_with()
    extra_info_to_idx_lookup.assert_not_called()
    extra_info_pgm_input_data.assert_not_called()
    assert len(converter.pp_output_data) == 1 and "res_line_3ph" in converter.pp_output_data
    assert len(converter.pgm_output_data) == 1 and "line" in converter.pgm_output_data
    assert len(result) == 1 and "res_line_3ph" in result


def test__serialize_data__invalid_output():
    # Arrange
    converter = PandaPowerConverter()

    # Act
    with pytest.raises(
        TypeError,
        match="Invalid output data dictionary supplied.",
    ):
        converter._serialize_data(data={"line": np.array([])}, extra_info=None)


@patch("power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._extra_info_to_idx_lookup")
@patch("power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._extra_info_to_pgm_input_data")
@patch("power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._create_output_data")
def test_serialize_data__extra_info(
    create_output_data_mock: MagicMock,
    extra_info_pgm_input_data_mock: MagicMock,
    extra_info_to_idx_lookup_mock: MagicMock,
):
    # Arrange
    converter = PandaPowerConverter()

    extra_info = MagicMock("extra_info")

    # Act
    converter._serialize_data(data={}, extra_info=extra_info)

    # Assert
    create_output_data_mock.assert_called_once_with()
    extra_info_pgm_input_data_mock.assert_called_once_with(extra_info)
    extra_info_to_idx_lookup_mock.assert_called_once_with(extra_info)


def test_extra_info_to_idx_lookup():
    # Arrange
    converter = PandaPowerConverter()
    extra_info = {
        0: {"id_reference": {"table": "bus", "index": 101}},
        1: {"id_reference": {"table": "bus", "index": 102}},
        2: {"id_reference": {"table": "bus", "index": 103}},
        3: {"id_reference": {"table": "load", "name": "const_current", "index": 201}, "node": 0},
        4: {"id_reference": {"table": "load", "name": "const_current", "index": 202}, "node": 1},
        5: {"id_reference": {"table": "load", "name": "const_current", "index": 203}, "node": 2},
        6: {"from_node": 0, "to_node": 1},
        7: {"from_node": 1, "to_node": 2},
    }

    # Act
    converter._extra_info_to_idx_lookup(extra_info=extra_info)

    # Assert
    pd.testing.assert_series_equal(converter.idx[("bus", None)], pd.Series([0, 1, 2], index=[101, 102, 103]))
    pd.testing.assert_series_equal(converter.idx_lookup[("bus", None)], pd.Series([101, 102, 103], index=[0, 1, 2]))

    pd.testing.assert_series_equal(
        converter.idx[("load", "const_current")], pd.Series([3, 4, 5], index=[201, 202, 203])
    )
    pd.testing.assert_series_equal(
        converter.idx_lookup[("load", "const_current")], pd.Series([201, 202, 203], index=[3, 4, 5])
    )


def test_extra_info_to_pgm_input_data():
    # Arrange
    converter = PandaPowerConverter()
    converter.pgm_output_data["node"] = initialize_array("sym_output", "node", 3)
    converter.pgm_output_data["line"] = initialize_array("sym_output", "line", 2)
    converter.pgm_output_data["node"]["id"] = [1, 2, 3]
    converter.pgm_output_data["line"]["id"] = [12, 23]
    extra_info = {
        12: {"pgm_input": {"from_node": 1, "to_node": 2, "i_n": 105.0}},
        23: {"pgm_input": {"from_node": 2, "to_node": 3, "i_n": 5.0}},
    }

    # Act
    converter._extra_info_to_pgm_input_data(extra_info=extra_info)

    # Assert
    assert "node" not in converter.pgm_input_data
    assert_struct_array_equal(
        converter.pgm_input_data["line"],
        [{"id": 12, "from_node": 1, "to_node": 2, "i_n": 105.0}, {"id": 23, "from_node": 2, "to_node": 3, "i_n": 5.0}],
    )


def test__extra_info_to_pp_input_data():
    converter = PandaPowerConverter()
    converter.pgm_output_data["transformer"] = initialize_array("sym_output", "transformer", 3)
    converter.pgm_output_data["transformer"]["id"] = [3, 4, 5]

    converter.idx_lookup[("trafo", None)] = pd.Series([201, 202, 203], index=[3, 4, 5])
    converter.idx[("trafo", None)] = pd.Series([3, 4, 5], index=[201, 202, 203])
    extra_info = {
        3: {"pp_input": {"df": 0.1}},
        4: {"pp_input": {"df": 0.2}},
        5: {"pp_input": {"df": 0.3}},
    }

    expected_trafo = pd.DataFrame({"df": [0.1, 0.2, 0.3]}, index=[201, 202, 203])

    converter._extra_info_to_pp_input_data(extra_info)

    pd.testing.assert_frame_equal(converter.pp_input_data["trafo"], expected_trafo)


def test__extra_info_to_pp_input_data__empty():
    converter = PandaPowerConverter()
    converter.pgm_output_data["line"] = initialize_array("sym_output", "line", 3)

    converter._extra_info_to_pp_input_data({})
    assert len(converter.pp_input_data) == 0


def test_create_input_data():
    # Arrange
    converter = MagicMock()

    # Act
    PandaPowerConverter._create_input_data(self=converter)  # type: ignore

    # Assert
    assert len(converter.method_calls) == 18
    converter._create_pgm_input_nodes.assert_called_once_with()
    converter._create_pgm_input_lines.assert_called_once_with()
    converter._create_pgm_input_sources.assert_called_once_with()
    converter._create_pgm_input_sym_loads.assert_called_once_with()
    converter._create_pgm_input_asym_loads.assert_called_once_with()
    converter._create_pgm_input_shunts.assert_called_once_with()
    converter._create_pgm_input_transformers.assert_called_once_with()
    converter._create_pgm_input_sym_gens.assert_called_once_with()
    converter._create_pgm_input_asym_gens.assert_called_once_with()
    converter._create_pgm_input_three_winding_transformers.assert_called_once_with()
    converter._create_pgm_input_links.assert_called_once_with()
    converter._create_pgm_input_storages.assert_called_once_with()
    converter._create_pgm_input_impedances.assert_called_once_with()
    converter._create_pgm_input_wards.assert_called_once_with()
    converter._create_pgm_input_xwards.assert_called_once_with()
    converter._create_pgm_input_motors.assert_called_once_with()
    converter._create_pgm_input_generators.assert_called_once_with()
    converter._create_pgm_input_dclines.assert_called_once_with()


@pytest.mark.parametrize(
    ("create_fn", "table"),
    [
        (PandaPowerConverter._create_pgm_input_nodes, "bus"),
        (PandaPowerConverter._create_pgm_input_lines, "line"),
        (PandaPowerConverter._create_pgm_input_sources, "ext_grid"),
        (PandaPowerConverter._create_pgm_input_shunts, "shunt"),
        (PandaPowerConverter._create_pgm_input_sym_gens, "sgen"),
        (PandaPowerConverter._create_pgm_input_sym_loads, "load"),
        (PandaPowerConverter._create_pgm_input_transformers, "trafo"),
        (PandaPowerConverter._create_pgm_input_three_winding_transformers, "trafo3w"),
        (PandaPowerConverter._create_pgm_input_links, "switch"),
    ],
)
@patch("power_grid_model_io.converters.pandapower_converter.initialize_array")
def test_create_pgm_input_object__empty(
    mock_init_array: MagicMock, create_fn: Callable[[PandaPowerConverter], None], table: str
):
    # Arrange
    converter = PandaPowerConverter()
    converter.pp_input_data[table] = pd.DataFrame()  # type: ignore

    # Act
    create_fn(converter)

    # Assert
    mock_init_array.assert_not_called()


@patch("power_grid_model_io.converters.pandapower_converter.initialize_array")
def test_create_pgm_input_nodes(mock_init_array: MagicMock, two_pp_objs: MockDf, converter):
    # Arrange
    converter.pp_input_data["bus"] = two_pp_objs

    # Act
    converter._create_pgm_input_nodes()

    # Assert

    # administration
    converter._generate_ids.assert_called_once_with("bus", two_pp_objs.index)

    # initialization
    mock_init_array.assert_called_once_with(data_type="input", component_type="node", shape=2)

    # retrieval
    converter._get_pp_attr.assert_any_call("bus", "vn_kv", expected_type="f8")
    assert len(converter._get_pp_attr.call_args_list) == 1

    # assignment
    pgm: MagicMock = mock_init_array.return_value.__setitem__
    pgm.assert_any_call("id", _generate_ids("bus", two_pp_objs.index))
    pgm.assert_any_call("u_rated", _get_pp_attr("bus", "vn_kv", expected_type="f8") * 1e3)
    assert len(pgm.call_args_list) == 2

    # result
    assert converter.pgm_input_data["node"] == mock_init_array.return_value


@patch("power_grid_model_io.converters.pandapower_converter.initialize_array")
def test_create_pgm_input_lines(mock_init_array: MagicMock, two_pp_objs, converter):
    # Arrange
    converter.pp_input_data["line"] = two_pp_objs

    # Act
    converter._create_pgm_input_lines()

    # Assert

    # administration
    converter.get_switch_states.assert_called_once_with("line")
    converter._generate_ids.assert_called_once_with("line", two_pp_objs.index)
    converter._get_pgm_ids.assert_any_call("bus", _get_pp_attr("line", "from_bus", expected_type="u4"))
    converter._get_pgm_ids.assert_any_call("bus", _get_pp_attr("line", "to_bus", expected_type="u4"))

    # initialization
    mock_init_array.assert_called_once_with(data_type="input", component_type="line", shape=2)

    # retrieval
    converter._get_pp_attr.assert_any_call("line", "from_bus", expected_type="u4")
    converter._get_pp_attr.assert_any_call("line", "to_bus", expected_type="u4")
    converter._get_pp_attr.assert_any_call("line", "in_service", expected_type="bool", default=True)
    converter._get_pp_attr.assert_any_call("line", "length_km", expected_type="f8")
    converter._get_pp_attr.assert_any_call("line", "parallel", expected_type="u4", default=1)
    converter._get_pp_attr.assert_any_call("line", "r_ohm_per_km", expected_type="f8")
    converter._get_pp_attr.assert_any_call("line", "x_ohm_per_km", expected_type="f8")
    converter._get_pp_attr.assert_any_call("line", "c_nf_per_km", expected_type="f8")
    converter._get_pp_attr.assert_any_call("line", "g_us_per_km", expected_type="f8", default=0)
    converter._get_pp_attr.assert_any_call("line", "max_i_ka", expected_type="f8", default=np.nan)
    converter._get_pp_attr.assert_any_call("line", "df", expected_type="f8", default=1)
    converter._get_pp_attr.assert_any_call("line", "r0_ohm_per_km", expected_type="f8", default=np.nan)
    converter._get_pp_attr.assert_any_call("line", "x0_ohm_per_km", expected_type="f8", default=np.nan)
    converter._get_pp_attr.assert_any_call("line", "c0_nf_per_km", expected_type="f8", default=np.nan)
    converter._get_pp_attr.assert_any_call("line", "g0_us_per_km", expected_type="f8", default=0)
    assert len(converter._get_pp_attr.call_args_list) == 15

    # assignment
    pgm: MagicMock = mock_init_array.return_value.__setitem__
    pgm.assert_any_call("id", _generate_ids("line", two_pp_objs.index))
    pgm.assert_any_call("from_node", _get_pgm_ids("bus", _get_pp_attr("line", "from_bus", expected_type="u4")))
    pgm.assert_any_call(
        "from_status",
        _get_pp_attr("line", "in_service", expected_type="bool", default=True) & get_switch_states("line")["from"],
    )
    pgm.assert_any_call("to_node", _get_pgm_ids("bus", _get_pp_attr("line", "to_bus", expected_type="u4")))
    pgm.assert_any_call(
        "to_status",
        _get_pp_attr("line", "in_service", expected_type="bool", default=True) & get_switch_states("line")["to"],
    )
    pgm.assert_any_call(
        "r1",
        _get_pp_attr("line", "r_ohm_per_km", expected_type="f8")
        * (
            _get_pp_attr("line", "length_km", expected_type="f8")
            / _get_pp_attr("line", "parallel", expected_type="u4", default=1)
        ),
    )
    pgm.assert_any_call(
        "x1",
        _get_pp_attr("line", "x_ohm_per_km", expected_type="f8")
        * (
            _get_pp_attr("line", "length_km", expected_type="f8")
            / _get_pp_attr("line", "parallel", expected_type="u4", default=1)
        ),
    )
    pgm.assert_any_call(
        "c1",
        _get_pp_attr("line", "c_nf_per_km", expected_type="f8")
        * _get_pp_attr("line", "length_km", expected_type="f8")
        * _get_pp_attr("line", "parallel", expected_type="u4", default=1)
        * 1e-9,
    )
    pgm.assert_any_call(
        "tan1",
        _get_pp_attr("line", "g_us_per_km", expected_type="f8", default=0)
        / _get_pp_attr("line", "c_nf_per_km", expected_type="f8")
        / (np.pi / 10),
    )
    pgm.assert_any_call(
        "i_n",
        (_get_pp_attr("line", "max_i_ka", expected_type="f8", default=np.nan) * 1e3)
        * _get_pp_attr("line", "df", expected_type="f8", default=1)
        * _get_pp_attr("line", "parallel", expected_type="u4", default=1),
    )
    pgm.assert_any_call("r0", ANY)
    pgm.assert_any_call("x0", ANY)
    pgm.assert_any_call("c0", ANY)
    pgm.assert_any_call("tan0", ANY)
    assert len(pgm.call_args_list) == 14

    # result
    assert converter.pgm_input_data["line"] == mock_init_array.return_value


@patch("power_grid_model_io.converters.pandapower_converter.initialize_array")
def test_create_pgm_input_sources(mock_init_array: MagicMock, two_pp_objs, converter):
    # Arrange
    converter.pp_input_data["ext_grid"] = two_pp_objs

    # Act
    converter._create_pgm_input_sources()

    # Assert

    # administration:
    converter._generate_ids.assert_called_once_with("ext_grid", two_pp_objs.index)

    # initialization
    mock_init_array.assert_called_once_with(data_type="input", component_type="source", shape=2)

    # retrieval:
    converter._get_pp_attr.assert_any_call("ext_grid", "bus", expected_type="u4")
    converter._get_pp_attr.assert_any_call("ext_grid", "vm_pu", expected_type="f8", default=1.0)
    converter._get_pp_attr.assert_any_call("ext_grid", "va_degree", expected_type="f8", default=0.0)
    converter._get_pp_attr.assert_any_call("ext_grid", "s_sc_max_mva", expected_type="f8", default=np.nan)
    converter._get_pp_attr.assert_any_call("ext_grid", "rx_max", expected_type="f8", default=np.nan)
    converter._get_pp_attr.assert_any_call("ext_grid", "in_service", expected_type="bool", default=True)
    converter._get_pp_attr.assert_any_call("ext_grid", "r0x0_max", expected_type="f8", default=np.nan)
    converter._get_pp_attr.assert_any_call("ext_grid", "x0x_max", expected_type="f8", default=np.nan)
    assert len(converter._get_pp_attr.call_args_list) == 8

    # assignment:
    pgm: MagicMock = mock_init_array.return_value.__setitem__
    pgm.assert_any_call("id", _generate_ids("ext_grid", two_pp_objs.index))
    pgm.assert_any_call("node", _get_pgm_ids("bus", _get_pp_attr("ext_grid", "bus", expected_type="u4")))
    pgm.assert_any_call("status", _get_pp_attr("ext_grid", "in_service", expected_type="bool", default=True))
    pgm.assert_any_call("u_ref", _get_pp_attr("ext_grid", "vm_pu", expected_type="f8", default=1.0))
    pgm.assert_any_call(
        "u_ref_angle", _get_pp_attr("ext_grid", "va_degree", expected_type="f8", default=0.0) * (np.pi / 180)
    )
    pgm.assert_any_call("sk", _get_pp_attr("ext_grid", "s_sc_max_mva", expected_type="f8", default=np.nan) * 1e6)
    pgm.assert_any_call("rx_ratio", _get_pp_attr("ext_grid", "rx_max", expected_type="f8", default=np.nan))
    assert len(pgm.call_args_list) == 7

    # result
    assert converter.pgm_input_data["source"] == mock_init_array.return_value


@pytest.mark.parametrize("kwargs", [{"r0x0_max": 0.5, "rx_max": 4}, {"x0x_max": 0.6}])
def test_create_pgm_input_sources__zero_sequence(kwargs) -> None:
    pp_net: pp.pandapowerNet = pp.create_empty_network()
    pp.create_bus(net=pp_net, vn_kv=1.0)
    pp.create_ext_grid(pp_net, 0, **kwargs)

    converter = PandaPowerConverter()
    converter.pp_input_data = {k: v for k, v in pp_net.items() if isinstance(v, pd.DataFrame)}
    converter.idx = {("bus", None): pd.Series([0], index=[0])}

    with patch("power_grid_model_io.converters.pandapower_converter.logger") as mock_logger:
        converter._create_pgm_input_sources()
        mock_logger.warning.assert_called_once()


@patch("power_grid_model_io.converters.pandapower_converter.initialize_array")
def test_create_pgm_input_sym_loads(mock_init_array: MagicMock, two_pp_objs, converter):
    # Arrange
    converter.pp_input_data["load"] = two_pp_objs
    pgm_attr = ["id", "node", "status", "p_specified", "q_specified", "type"]
    pgm = {attr: MagicMock() for attr in pgm_attr}
    mock_init_array.return_value = pgm
    slices = [slice(None, 2), slice(2, 4), slice(-2, None)]
    assert slices[0].indices(3 * 2) == (0, 2, 1)
    assert slices[1].indices(3 * 2) == (2, 4, 1)
    assert slices[2].indices(3 * 2) == (4, 6, 1)

    # Act
    converter._create_pgm_input_sym_loads()

    # Assert

    # administration:

    # initialization
    mock_init_array.assert_called_once_with(data_type="input", component_type="sym_load", shape=3 * 2)

    # retrieval:
    converter._get_pp_attr.assert_any_call("load", "bus", expected_type="u4")
    converter._get_pp_attr.assert_any_call("load", "p_mw", expected_type="f8", default=0.0)
    converter._get_pp_attr.assert_any_call("load", "q_mvar", expected_type="f8", default=0.0)
    converter._get_pp_attr.assert_any_call("load", "const_z_percent", expected_type="f8", default=0)
    converter._get_pp_attr.assert_any_call("load", "const_i_percent", expected_type="f8", default=0)
    converter._get_pp_attr.assert_any_call("load", "scaling", expected_type="f8", default=1)
    converter._get_pp_attr.assert_any_call("load", "in_service", expected_type="bool", default=True)
    converter._get_pp_attr.assert_any_call("load", "type", expected_type="O", default=None)
    assert len(converter._get_pp_attr.call_args_list) == 8

    # assignment:
    for attr in pgm_attr:
        for s in slices:
            pgm[attr].__setitem__.assert_any_call(s, ANY)
        assert len(pgm[attr].__setitem__.call_args_list) == len(slices)

    # result
    assert converter.pgm_input_data["sym_load"] == pgm


@patch("power_grid_model_io.converters.pandapower_converter.initialize_array")
def test_create_pgm_input_asym_loads(mock_init_array: MagicMock, two_pp_objs, converter):
    # Arrange
    converter.pp_input_data["asymmetric_load"] = two_pp_objs

    # Act
    converter._create_pgm_input_asym_loads()

    # Assert

    # administration:
    converter._generate_ids.assert_called_once_with("asymmetric_load", two_pp_objs.index)

    # initialization
    mock_init_array.assert_called_once_with(data_type="input", component_type="asym_load", shape=2)

    # retrieval:
    converter._get_pp_attr.assert_any_call("asymmetric_load", "bus", expected_type="u4")
    converter._get_pp_attr.assert_any_call("asymmetric_load", "p_a_mw", expected_type="f8")
    converter._get_pp_attr.assert_any_call("asymmetric_load", "p_b_mw", expected_type="f8")
    converter._get_pp_attr.assert_any_call("asymmetric_load", "p_c_mw", expected_type="f8")
    converter._get_pp_attr.assert_any_call("asymmetric_load", "q_a_mvar", expected_type="f8")
    converter._get_pp_attr.assert_any_call("asymmetric_load", "q_b_mvar", expected_type="f8")
    converter._get_pp_attr.assert_any_call("asymmetric_load", "q_c_mvar", expected_type="f8")
    converter._get_pp_attr.assert_any_call("asymmetric_load", "scaling", expected_type="f8")
    converter._get_pp_attr.assert_any_call("asymmetric_load", "in_service", expected_type="bool", default=True)
    converter._get_pp_attr.assert_any_call("asymmetric_load", "type", expected_type="O", default=None)
    assert len(converter._get_pp_attr.call_args_list) == 10

    # assignment:
    pgm: MagicMock = mock_init_array.return_value.__setitem__
    pgm.assert_any_call("id", _generate_ids("asymmetric_load", two_pp_objs.index))
    pgm.assert_any_call("node", _get_pgm_ids("bus", _get_pp_attr("asymmetric_load", "bus", expected_type="u4")))
    pgm.assert_any_call("status", _get_pp_attr("asymmetric_load", "in_service", expected_type="bool", default=True))
    pgm.assert_any_call("p_specified", ANY)
    pgm.assert_any_call("q_specified", ANY)
    assert len(pgm.call_args_list) == 6
    # result
    assert converter.pgm_input_data["asym_load"] == mock_init_array.return_value


def test_create_pgm_input_sym_loads__delta() -> None:
    # Arrange
    pp_net: pp.pandapowerNet = pp.create_empty_network()
    pp.create_bus(net=pp_net, vn_kv=0.0)
    pp.create_load(pp_net, 0, 0, type="delta")

    converter = PandaPowerConverter()
    converter.pp_input_data = {k: v for k, v in pp_net.items() if isinstance(v, pd.DataFrame)}

    # Act/Assert
    with pytest.raises(
        NotImplementedError, match="Delta loads are not implemented, only wye loads are supported in PGM."
    ):
        converter._create_pgm_input_sym_loads()


def test_create_pgm_input_asym_loads__delta() -> None:
    # Arrange
    pp_net: pp.pandapowerNet = pp.create_empty_network()
    pp.create_bus(net=pp_net, vn_kv=0.0)
    pp.create_asymmetric_load(pp_net, 0, type="delta")

    converter = PandaPowerConverter()
    converter.pp_input_data = {k: v for k, v in pp_net.items() if isinstance(v, pd.DataFrame)}

    # Act/Assert
    with pytest.raises(
        NotImplementedError, match="Delta loads are not implemented, only wye loads are supported in PGM."
    ):
        converter._create_pgm_input_asym_loads()


def test_create_pgm_input_transformers__tap_dependent_impedance() -> None:
    # Arrange
    pp_net: pp.pandapowerNet = pp.create_empty_network()
    pp.create_bus(net=pp_net, vn_kv=0.0)
    args = [0, 0, 0, 0, 0, 0, 0, 0, 0]
    pp.create_transformer_from_parameters(pp_net, *args, tap_dependent_impedance=True)

    converter = PandaPowerConverter()
    converter.pp_input_data = {k: v for k, v in pp_net.items() if isinstance(v, pd.DataFrame)}

    # Act/Assert
    with pytest.raises(RuntimeError, match="not supported"):
        converter._create_pgm_input_transformers()


@patch("power_grid_model_io.converters.pandapower_converter.initialize_array")
def test_create_pgm_input_shunts(mock_init_array: MagicMock, two_pp_objs, converter):
    # Arrange
    converter.pp_input_data["shunt"] = two_pp_objs

    # Act
    converter._create_pgm_input_shunts()

    # Assert

    # administration:
    converter._generate_ids.assert_called_once_with("shunt", two_pp_objs.index)

    # initialization
    mock_init_array.assert_called_once_with(data_type="input", component_type="shunt", shape=2)

    # retrieval:
    converter._get_pp_attr.assert_any_call("shunt", "bus", expected_type="u4")
    converter._get_pp_attr.assert_any_call("shunt", "p_mw", expected_type="f8")
    converter._get_pp_attr.assert_any_call("shunt", "q_mvar", expected_type="f8")
    converter._get_pp_attr.assert_any_call("shunt", "vn_kv", expected_type="f8")
    converter._get_pp_attr.assert_any_call("shunt", "step", expected_type="u4", default=1)
    converter._get_pp_attr.assert_any_call("shunt", "in_service", expected_type="bool", default=True)

    assert len(converter._get_pp_attr.call_args_list) == 6

    # assignment:
    pgm: MagicMock = mock_init_array.return_value.__setitem__
    pgm.assert_any_call("id", _generate_ids("shunt", two_pp_objs.index))
    pgm.assert_any_call("node", _get_pgm_ids("bus", _get_pp_attr("shunt", "bus", expected_type="u4")))
    pgm.assert_any_call("status", _get_pp_attr("shunt", "in_service", expected_type="bool", default=True))
    pgm.assert_any_call(
        "g1",
        _get_pp_attr("shunt", "p_mw", expected_type="f8")
        * _get_pp_attr("shunt", "step", expected_type="u4", default=1)
        / _get_pp_attr("shunt", "vn_kv", expected_type="f8")
        / _get_pp_attr("shunt", "vn_kv", expected_type="f8"),
    )
    pgm.assert_any_call(
        "b1",
        -_get_pp_attr("shunt", "q_mvar", expected_type="f8")
        * _get_pp_attr("shunt", "step", expected_type="u4", default=1)
        / _get_pp_attr("shunt", "vn_kv", expected_type="f8")
        / _get_pp_attr("shunt", "vn_kv", expected_type="f8"),
    )
    pgm.assert_any_call("g0", ANY)
    pgm.assert_any_call("b0", ANY)

    assert len(pgm.call_args_list) == 7

    # result
    assert converter.pgm_input_data["shunt"] == mock_init_array.return_value


@patch("power_grid_model_io.converters.pandapower_converter.initialize_array")
@patch("power_grid_model_io.converters.pandapower_converter.np.round", new=lambda x: x)
@patch("power_grid_model_io.converters.pandapower_converter.np.divide", new=lambda x, _, **kwargs: x)
@patch("power_grid_model_io.converters.pandapower_converter.np.bitwise_and", new=lambda x, _: x)
@patch("power_grid_model_io.converters.pandapower_converter.np.logical_and", new=lambda x, _: x)
@patch("power_grid_model_io.converters.pandapower_converter.np.allclose", new=lambda x, _: x)
@patch("power_grid_model_io.converters.pandapower_converter.np.isnan", new=lambda x: x)
def test_create_pgm_input_transformers(mock_init_array: MagicMock, two_pp_objs, converter):
    # Arrange
    converter.pp_input_data["trafo"] = two_pp_objs

    # Act
    converter._create_pgm_input_transformers()
    # Assert

    # administration:
    converter.get_switch_states.assert_called_once_with("trafo")
    converter._generate_ids.assert_called_once_with("trafo", two_pp_objs.index)
    converter._get_pgm_ids.assert_any_call("bus", _get_pp_attr("trafo", "hv_bus", expected_type="u4"))
    converter._get_pgm_ids.assert_any_call("bus", _get_pp_attr("trafo", "lv_bus", expected_type="u4"))
    converter._get_tap_size.assert_called_once_with(two_pp_objs)
    converter._get_transformer_tap_side.assert_called_once_with(
        _get_pp_attr("trafo", "tap_side", expected_type="O", default=None)
    )

    # initialization
    mock_init_array.assert_called_once_with(data_type="input", component_type="transformer", shape=2)

    # retrieval:
    converter._get_pp_attr.assert_any_call("trafo", "hv_bus", expected_type="u4")
    converter._get_pp_attr.assert_any_call("trafo", "lv_bus", expected_type="u4")
    converter._get_pp_attr.assert_any_call("trafo", "sn_mva", expected_type="f8")
    converter._get_pp_attr.assert_any_call("trafo", "vn_hv_kv", expected_type="f8")
    converter._get_pp_attr.assert_any_call("trafo", "vn_lv_kv", expected_type="f8")
    converter._get_pp_attr.assert_any_call("trafo", "vk_percent", expected_type="f8")
    converter._get_pp_attr.assert_any_call("trafo", "vkr_percent", expected_type="f8")
    converter._get_pp_attr.assert_any_call("trafo", "pfe_kw", expected_type="f8")
    converter._get_pp_attr.assert_any_call("trafo", "i0_percent", expected_type="f8")
    converter._get_pp_attr.assert_any_call("trafo", "shift_degree", expected_type="f8", default=0.0)
    converter._get_pp_attr.assert_any_call("trafo", "tap_side", expected_type="O", default=None)
    converter._get_pp_attr.assert_any_call("trafo", "tap_neutral", expected_type="f8", default=np.nan)
    converter._get_pp_attr.assert_any_call("trafo", "tap_min", expected_type="i4", default=0)
    converter._get_pp_attr.assert_any_call("trafo", "tap_max", expected_type="i4", default=0)
    converter._get_pp_attr.assert_any_call("trafo", "tap_pos", expected_type="f8", default=np.nan)
    converter._get_pp_attr.assert_any_call("trafo", "parallel", expected_type="u4", default=1)
    converter._get_pp_attr.assert_any_call("trafo", "in_service", expected_type="bool", default=True)
    converter._get_pp_attr.assert_any_call("trafo", "vk0_percent", expected_type="f8", default=np.nan)
    converter._get_pp_attr.assert_any_call("trafo", "vkr0_percent", expected_type="f8", default=np.nan)
    converter._get_pp_attr.assert_any_call("trafo", "mag0_percent", expected_type="f8", default=np.nan)
    converter._get_pp_attr.assert_any_call("trafo", "mag0_rx", expected_type="f8", default=np.nan)
    converter._get_pp_attr.assert_any_call("trafo", "si0_hv_partial", expected_type="f8", default=np.nan)
    # converter._get_pp_attr.assert_any_call('trafo', 'df', expected_type='f8')  #TODO add df in output conversions
    assert len(converter._get_pp_attr.call_args_list) == 22

    # assignment:
    pgm: MagicMock = mock_init_array.return_value.__setitem__
    pgm.assert_any_call("id", _generate_ids("trafo", two_pp_objs.index))
    pgm.assert_any_call("from_node", _get_pgm_ids("bus", _get_pp_attr("trafo", "hv_bus", expected_type="u4")))
    pgm.assert_any_call("to_node", _get_pgm_ids("bus", _get_pp_attr("trafo", "lv_bus", expected_type="u4")))
    pgm.assert_any_call("from_status", ANY)
    pgm.assert_any_call("to_status", ANY)
    pgm.assert_any_call("u1", ANY)
    pgm.assert_any_call("u2", ANY)
    pgm.assert_any_call("sn", ANY)
    pgm.assert_any_call("uk", ANY)
    pgm.assert_any_call("pk", ANY)
    pgm.assert_any_call("i0", ANY)
    pgm.assert_any_call("p0", ANY)
    pgm.assert_any_call("winding_from", ANY)
    pgm.assert_any_call("winding_to", ANY)
    pgm.assert_any_call("clock", ANY)
    pgm.assert_any_call("tap_side", ANY)
    pgm.assert_any_call("tap_pos", ANY)
    pgm.assert_any_call("tap_min", ANY)
    pgm.assert_any_call("tap_max", ANY)
    pgm.assert_any_call("tap_nom", ANY)
    pgm.assert_any_call("tap_size", ANY)

    assert len(pgm.call_args_list) == 21

    # result
    assert converter.pgm_input_data["transformer"] == mock_init_array.return_value


@patch(
    "power_grid_model_io.converters.pandapower_converter.PandaPowerConverter.get_switch_states",
    new=MagicMock(return_value=pd.DataFrame({"from": [True], "to": [True]})),
)
@patch(
    "power_grid_model_io.converters.pandapower_converter.PandaPowerConverter.get_trafo_winding_types",
    new=MagicMock(
        return_value=pd.DataFrame(
            {
                "winding_from": [0, 0, 0, 0, 0, 0, WindingType.delta, WindingType.wye_n, 0, 0],
                "winding_to": [0, 0, 0, 0, 0, 0, WindingType.wye_n, WindingType.wye_n, 0, 0],
            }
        )
    ),
)
@patch(
    "power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._generate_ids",
    new=MagicMock(return_value=np.arange(1)),
)
@patch(
    "power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._get_pgm_ids",
    new=MagicMock(return_value=pd.Series([0])),
)
def test_create_pgm_input_transformers__default() -> None:
    # Arrange
    pp_net: pp.pandapowerNet = pp.create_empty_network()
    pp.create_bus(net=pp_net, vn_kv=0.0)
    args = [0, 0, 0, 0, 0, 0, 0, 0, 0]
    pp.create_transformer_from_parameters(
        pp_net, *args, tap_neutral=12.0, tap_pos=34.0, tap_step_percent=1, tap_side="hv"
    )
    pp.create_transformer_from_parameters(
        pp_net, *args, tap_neutral=12.0, tap_pos=34.0, tap_step_percent=1, tap_side="lv"
    )
    pp.create_transformer_from_parameters(
        pp_net, *args, tap_neutral=12.0, tap_pos=34.0, tap_step_percent=1, tap_side=None
    )
    tap_pos_trafo = pp.create_transformer_from_parameters(pp_net, *args, tap_neutral=12.0, tap_size=1, tap_side="hv")
    pp_net["trafo"].loc[tap_pos_trafo, "tap_pos"] = np.nan
    pp.create_transformer_from_parameters(pp_net, *args, tap_neutral=np.nan, tap_pos=34.0, tap_side="hv")
    pp.create_transformer_from_parameters(
        pp_net, *args, tap_neutral=12, tap_step_percent=np.nan, tap_pos=34.0, tap_side="hv"
    )
    pp.create_transformer_from_parameters(pp_net, *args, vector_group=None, shift_degree=30)
    pp.create_transformer_from_parameters(pp_net, *args, vector_group=None, shift_degree=60)
    pp.create_transformer_from_parameters(pp_net, *args, vector_group=None, shift_degree=59)
    pp.create_transformer_from_parameters(pp_net, *args, vector_group=None, shift_degree=61)

    converter = PandaPowerConverter()
    converter.pp_input_data = {k: v for k, v in pp_net.items() if isinstance(v, pd.DataFrame)}

    # Act
    converter._create_pgm_input_transformers()
    result = converter.pgm_input_data["transformer"]

    # Assert
    assert result[0]["tap_side"] == BranchSide.from_side.value
    assert result[1]["tap_side"] == BranchSide.to_side.value
    assert result[2]["tap_side"] == BranchSide.from_side.value
    assert result[3]["tap_side"] == BranchSide.from_side.value
    assert result[4]["tap_side"] == BranchSide.from_side.value
    assert result[0]["tap_pos"] == 34.0 != result[0]["tap_nom"]
    assert result[1]["tap_pos"] == 34.0 != result[1]["tap_nom"]
    assert result[2]["tap_pos"] == 0.0 == result[2]["tap_nom"]
    assert result[3]["tap_pos"] == 0.0 == result[3]["tap_nom"]
    assert result[4]["tap_pos"] == 0.0 == result[4]["tap_nom"]
    assert result[5]["tap_size"] == 0.0

    assert result[6]["winding_from"] == WindingType.delta
    assert result[6]["winding_to"] == WindingType.wye_n
    assert result[7]["winding_from"] == WindingType.wye_n
    assert result[7]["winding_to"] == WindingType.wye_n

    assert result[8]["clock"] == 2
    assert result[9]["clock"] == 2


@patch("power_grid_model_io.converters.pandapower_converter.initialize_array")
def test_create_pgm_input_sym_gens(mock_init_array: MagicMock, two_pp_objs, converter):
    # Arrange
    converter.pp_input_data["sgen"] = two_pp_objs

    # Act
    converter._create_pgm_input_sym_gens()

    # Assert

    # administration:
    converter._generate_ids.assert_called_once_with("sgen", two_pp_objs.index)

    # initialization
    mock_init_array.assert_called_once_with(data_type="input", component_type="sym_gen", shape=2)

    # retrieval:
    converter._get_pp_attr.assert_any_call("sgen", "bus", expected_type="i8")
    converter._get_pp_attr.assert_any_call("sgen", "p_mw", expected_type="f8")
    converter._get_pp_attr.assert_any_call("sgen", "q_mvar", expected_type="f8", default=0.0)
    converter._get_pp_attr.assert_any_call("sgen", "scaling", expected_type="f8", default=1.0)
    converter._get_pp_attr.assert_any_call("sgen", "in_service", expected_type="bool", default=True)
    assert len(converter._get_pp_attr.call_args_list) == 5

    # assignment:
    pgm: MagicMock = mock_init_array.return_value.__setitem__
    pgm.assert_any_call("id", _generate_ids("sgen", two_pp_objs.index))
    pgm.assert_any_call("node", _get_pgm_ids("bus", _get_pp_attr("sgen", "bus", expected_type="i8")))
    pgm.assert_any_call("status", _get_pp_attr("sgen", "in_service", expected_type="bool", default=True))
    pgm.assert_any_call("type", LoadGenType.const_power)
    pgm.assert_any_call(
        "p_specified",
        _get_pp_attr("sgen", "p_mw", expected_type="f8")
        * _get_pp_attr("sgen", "scaling", expected_type="f8", default=1.0)
        * 1e6,
    )
    pgm.assert_any_call(
        "q_specified",
        _get_pp_attr("sgen", "q_mvar", expected_type="f8", default=0.0)
        * _get_pp_attr("sgen", "scaling", expected_type="f8", default=1.0)
        * 1e6,
    )
    assert len(pgm.call_args_list) == 6

    # result
    assert converter.pgm_input_data["sym_gen"] == mock_init_array.return_value


@pytest.mark.parametrize(
    "kwargs",
    [{"vk0_percent": 2}, {"vkr0_percent": 1}, {"mag0_percent": 5}, {"mag0_rx": 0.2}, {"si0_hv_partial": 0.3}],
)
@patch(
    "power_grid_model_io.converters.pandapower_converter.PandaPowerConverter.get_switch_states",
    new=MagicMock(return_value=pd.DataFrame({"from": [True], "to": [True]})),
)
@patch(
    "power_grid_model_io.converters.pandapower_converter.PandaPowerConverter.get_trafo_winding_types",
    new=MagicMock(return_value=pd.DataFrame({"winding_from": [0], "winding_to": [0]})),
)
@patch(
    "power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._generate_ids",
    new=MagicMock(return_value=np.arange(1)),
)
@patch(
    "power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._get_pgm_ids",
    new=MagicMock(return_value=pd.Series([0])),
)
def test_create_pgm_input_transformers__zero_sequence(kwargs) -> None:
    # Arrange
    pp_net: pp.pandapowerNet = pp.create_empty_network()
    pp.create_bus(net=pp_net, vn_kv=0.0)
    args = [0, 0, 0, 0, 0, 0, 0, 0, 0]
    pp.create_transformer_from_parameters(pp_net, *args, **kwargs)

    converter = PandaPowerConverter()
    converter.pp_input_data = {k: v for k, v in pp_net.items() if isinstance(v, pd.DataFrame)}

    with patch("power_grid_model_io.converters.pandapower_converter.logger") as mock_logger:
        converter._create_pgm_input_transformers()
        mock_logger.warning.assert_called_once()


@patch("power_grid_model_io.converters.pandapower_converter.initialize_array")
def test_create_pgm_input_asym_gens(mock_init_array: MagicMock, two_pp_objs, converter):
    # Arrange
    converter.pp_input_data["asymmetric_sgen"] = two_pp_objs

    # Act
    converter._create_pgm_input_asym_gens()

    # Assert

    # administration:
    converter._generate_ids.assert_called_once_with("asymmetric_sgen", two_pp_objs.index)

    # initialization
    mock_init_array.assert_called_once_with(data_type="input", component_type="asym_gen", shape=2)

    # retrieval:
    converter._get_pp_attr.assert_any_call("asymmetric_sgen", "bus", expected_type="i8")
    converter._get_pp_attr.assert_any_call("asymmetric_sgen", "p_a_mw", expected_type="f8")
    converter._get_pp_attr.assert_any_call("asymmetric_sgen", "p_b_mw", expected_type="f8")
    converter._get_pp_attr.assert_any_call("asymmetric_sgen", "p_c_mw", expected_type="f8")
    converter._get_pp_attr.assert_any_call("asymmetric_sgen", "q_a_mvar", expected_type="f8")
    converter._get_pp_attr.assert_any_call("asymmetric_sgen", "q_b_mvar", expected_type="f8")
    converter._get_pp_attr.assert_any_call("asymmetric_sgen", "q_c_mvar", expected_type="f8")
    converter._get_pp_attr.assert_any_call("asymmetric_sgen", "scaling", expected_type="f8")
    converter._get_pp_attr.assert_any_call("asymmetric_sgen", "in_service", expected_type="bool", default=True)
    assert len(converter._get_pp_attr.call_args_list) == 9

    # assignment:
    pgm: MagicMock = mock_init_array.return_value.__setitem__
    pgm.assert_any_call("id", _generate_ids("asymmetric_sgen", two_pp_objs.index))
    pgm.assert_any_call("node", _get_pgm_ids("bus", _get_pp_attr("asymmetric_sgen", "bus", expected_type="i8")))
    pgm.assert_any_call("status", _get_pp_attr("asymmetric_sgen", "in_service", expected_type="bool", default=True))
    pgm.assert_any_call("p_specified", ANY)
    pgm.assert_any_call("q_specified", ANY)
    pgm.assert_any_call("type", LoadGenType.const_power)
    assert len(pgm.call_args_list) == 6

    # result
    assert converter.pgm_input_data["asym_gen"] == mock_init_array.return_value


@patch("power_grid_model_io.converters.pandapower_converter.initialize_array")
@patch("power_grid_model_io.converters.pandapower_converter.np.round", new=lambda x: x)
def test_create_pgm_input_three_winding_transformers(mock_init_array: MagicMock, two_pp_objs, converter):
    # Arrange
    converter.pp_input_data["trafo3w"] = two_pp_objs

    # Act
    converter._create_pgm_input_three_winding_transformers()

    # Assert

    # administration:
    converter.get_trafo3w_switch_states.assert_called_once_with(two_pp_objs)
    converter._generate_ids.assert_called_once_with("trafo3w", two_pp_objs.index)
    converter._get_pgm_ids.assert_any_call("bus", _get_pp_attr("trafo3w", "hv_bus", expected_type="u4"))
    converter._get_pgm_ids.assert_any_call("bus", _get_pp_attr("trafo3w", "mv_bus", expected_type="u4"))
    converter._get_pgm_ids.assert_any_call("bus", _get_pp_attr("trafo3w", "lv_bus", expected_type="u4"))
    converter._get_3wtransformer_tap_size.assert_called_once_with(two_pp_objs)
    converter._get_3wtransformer_tap_side.assert_called_once_with(
        _get_pp_attr("trafo3w", "tap_side", expected_type="O", default=None)
    )

    # initialization
    mock_init_array.assert_called_once_with(data_type="input", component_type="three_winding_transformer", shape=2)

    # retrieval:
    converter._get_pp_attr.assert_any_call("trafo3w", "hv_bus", expected_type="u4")
    converter._get_pp_attr.assert_any_call("trafo3w", "mv_bus", expected_type="u4")
    converter._get_pp_attr.assert_any_call("trafo3w", "lv_bus", expected_type="u4")
    converter._get_pp_attr.assert_any_call("trafo3w", "vn_hv_kv", expected_type="f8")
    converter._get_pp_attr.assert_any_call("trafo3w", "vn_mv_kv", expected_type="f8")
    converter._get_pp_attr.assert_any_call("trafo3w", "vn_lv_kv", expected_type="f8")
    converter._get_pp_attr.assert_any_call("trafo3w", "sn_hv_mva", expected_type="f8")
    converter._get_pp_attr.assert_any_call("trafo3w", "sn_mv_mva", expected_type="f8")
    converter._get_pp_attr.assert_any_call("trafo3w", "sn_lv_mva", expected_type="f8")
    converter._get_pp_attr.assert_any_call("trafo3w", "vk_hv_percent", expected_type="f8")
    converter._get_pp_attr.assert_any_call("trafo3w", "vk_mv_percent", expected_type="f8")
    converter._get_pp_attr.assert_any_call("trafo3w", "vk_lv_percent", expected_type="f8")
    converter._get_pp_attr.assert_any_call("trafo3w", "vkr_hv_percent", expected_type="f8")
    converter._get_pp_attr.assert_any_call("trafo3w", "vkr_mv_percent", expected_type="f8")
    converter._get_pp_attr.assert_any_call("trafo3w", "vkr_lv_percent", expected_type="f8")
    converter._get_pp_attr.assert_any_call("trafo3w", "pfe_kw", expected_type="f8")
    converter._get_pp_attr.assert_any_call("trafo3w", "i0_percent", expected_type="f8")
    converter._get_pp_attr.assert_any_call("trafo3w", "shift_mv_degree", expected_type="f8", default=0.0)
    converter._get_pp_attr.assert_any_call("trafo3w", "shift_lv_degree", expected_type="f8", default=0.0)
    converter._get_pp_attr.assert_any_call("trafo3w", "tap_side", expected_type="O", default=None)
    converter._get_pp_attr.assert_any_call("trafo3w", "tap_neutral", expected_type="f8", default=np.nan)
    converter._get_pp_attr.assert_any_call("trafo3w", "tap_min", expected_type="i4", default=0)
    converter._get_pp_attr.assert_any_call("trafo3w", "tap_max", expected_type="i4", default=0)
    converter._get_pp_attr.assert_any_call("trafo3w", "tap_pos", expected_type="f8", default=np.nan)
    converter._get_pp_attr.assert_any_call("trafo3w", "in_service", expected_type="bool", default=True)
    converter._get_pp_attr.assert_any_call("trafo3w", "vk0_hv_percent", expected_type="f8", default=np.nan)
    converter._get_pp_attr.assert_any_call("trafo3w", "vkr0_hv_percent", expected_type="f8", default=np.nan)
    converter._get_pp_attr.assert_any_call("trafo3w", "vk0_mv_percent", expected_type="f8", default=np.nan)
    converter._get_pp_attr.assert_any_call("trafo3w", "vkr0_mv_percent", expected_type="f8", default=np.nan)
    converter._get_pp_attr.assert_any_call("trafo3w", "vk0_lv_percent", expected_type="f8", default=np.nan)
    converter._get_pp_attr.assert_any_call("trafo3w", "vkr0_lv_percent", expected_type="f8", default=np.nan)
    assert len(converter._get_pp_attr.call_args_list) == 31

    # assignment:
    pgm: MagicMock = mock_init_array.return_value.__setitem__
    pgm.assert_any_call("id", _generate_ids("trafo3w", two_pp_objs.index))
    pgm.assert_any_call("node_1", _get_pgm_ids("bus", _get_pp_attr("trafo3w", "hv_bus", expected_type="u4")))
    pgm.assert_any_call("node_2", _get_pgm_ids("bus", _get_pp_attr("trafo3w", "mv_bus", expected_type="u4")))
    pgm.assert_any_call("node_3", _get_pgm_ids("bus", _get_pp_attr("trafo3w", "lv_bus", expected_type="u4")))
    pgm.assert_any_call("status_1", ANY)
    pgm.assert_any_call("status_2", ANY)
    pgm.assert_any_call("status_3", ANY)
    pgm.assert_any_call("u1", ANY)
    pgm.assert_any_call("u2", ANY)
    pgm.assert_any_call("u3", ANY)
    pgm.assert_any_call("sn_1", ANY)
    pgm.assert_any_call("sn_2", ANY)
    pgm.assert_any_call("sn_3", ANY)
    pgm.assert_any_call("uk_12", ANY)
    pgm.assert_any_call("uk_13", ANY)
    pgm.assert_any_call("uk_23", ANY)
    pgm.assert_any_call("pk_12", ANY)
    pgm.assert_any_call("pk_13", ANY)
    pgm.assert_any_call("pk_23", ANY)
    pgm.assert_any_call("i0", ANY)
    pgm.assert_any_call("p0", ANY)
    pgm.assert_any_call("winding_1", ANY)
    pgm.assert_any_call("winding_2", ANY)
    pgm.assert_any_call("winding_3", ANY)
    pgm.assert_any_call("clock_12", ANY)
    pgm.assert_any_call("clock_13", ANY)
    pgm.assert_any_call("tap_side", ANY)
    pgm.assert_any_call("tap_pos", ANY)
    pgm.assert_any_call("tap_min", ANY)
    pgm.assert_any_call("tap_max", ANY)
    pgm.assert_any_call("tap_nom", ANY)
    pgm.assert_any_call("tap_size", ANY)
    assert len(pgm.call_args_list) == 32

    # result
    assert converter.pgm_input_data["three_winding_transformer"] == mock_init_array.return_value


@patch(
    "power_grid_model_io.converters.pandapower_converter.PandaPowerConverter.get_trafo3w_switch_states",
    new=MagicMock(return_value=pd.DataFrame({"side_1": [True], "side_2": [True], "side_3": [True]})),
)
@patch(
    "power_grid_model_io.converters.pandapower_converter.PandaPowerConverter.get_trafo3w_winding_types",
    new=MagicMock(
        return_value=pd.DataFrame(
            {
                "winding_1": [
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    WindingType.wye_n,
                    WindingType.wye_n,
                    WindingType.wye_n,
                    WindingType.wye_n,
                    0,
                    0,
                ],
                "winding_2": [
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    WindingType.delta,
                    WindingType.wye_n,
                    WindingType.wye_n,
                    WindingType.delta,
                    0,
                    0,
                ],
                "winding_3": [
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    WindingType.delta,
                    WindingType.wye_n,
                    WindingType.delta,
                    WindingType.wye_n,
                    0,
                    0,
                ],
            }
        )
    ),
)
@patch(
    "power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._generate_ids",
    new=MagicMock(return_value=np.arange(1)),
)
@patch(
    "power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._get_pgm_ids",
    new=MagicMock(return_value=pd.Series([0])),
)
def test_create_pgm_input_transformers3w__default() -> None:
    # Arrange
    pp_net: pp.pandapowerNet = pp.create_empty_network()
    pp.create_bus(net=pp_net, vn_kv=0.0)
    args = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    pp.create_transformer3w_from_parameters(
        pp_net, *args, tap_neutral=12.0, tap_pos=34.0, tap_step_percent=1, tap_side="hv"
    )
    pp.create_transformer3w_from_parameters(
        pp_net, *args, tap_neutral=12.0, tap_pos=34.0, tap_step_percent=1, tap_side="mv"
    )
    pp.create_transformer3w_from_parameters(
        pp_net, *args, tap_neutral=12.0, tap_pos=34.0, tap_step_percent=1, tap_side="lv"
    )
    pp.create_transformer3w_from_parameters(
        pp_net, *args, tap_neutral=12.0, tap_pos=34.0, tap_step_percent=1, tap_side=None
    )
    pp.create_transformer3w_from_parameters(
        pp_net, *args, tap_neutral=np.nan, tap_pos=34.0, tap_step_percent=1, tap_side="hv"
    )
    nan_trafo = pp.create_transformer3w_from_parameters(
        pp_net, *args, tap_neutral=12.0, tap_step_percent=1, tap_pos=np.nan, tap_side="hv"
    )
    pp_net["trafo3w"].loc[nan_trafo, "tap_pos"] = np.nan
    pp.create_transformer3w_from_parameters(
        pp_net, *args, tap_neutral=12.0, tap_pos=34.0, tap_step_percent=np.nan, tap_side="hv"
    )
    pp.create_transformer3w_from_parameters(
        pp_net,
        *args,
        vector_group=None,
        shift_mv_degree=30,
        shift_lv_degree=30,
    )
    pp.create_transformer3w_from_parameters(
        pp_net,
        *args,
        vector_group=None,
        shift_mv_degree=60,
        shift_lv_degree=60,
    )
    pp.create_transformer3w_from_parameters(
        pp_net,
        *args,
        vector_group=None,
        shift_mv_degree=60,
        shift_lv_degree=30,
    )
    pp.create_transformer3w_from_parameters(
        pp_net,
        *args,
        vector_group=None,
        shift_mv_degree=30,
        shift_lv_degree=60,
    )
    pp.create_transformer3w_from_parameters(
        pp_net,
        *args,
        vector_group=None,
        shift_mv_degree=58,
        shift_lv_degree=62,
    )
    pp.create_transformer3w_from_parameters(
        pp_net,
        *args,
        vector_group=None,
        shift_mv_degree=29,
        shift_lv_degree=31,
    )

    converter = PandaPowerConverter()
    converter.pp_input_data = {k: v for k, v in pp_net.items() if isinstance(v, pd.DataFrame)}

    # Act
    converter._create_pgm_input_three_winding_transformers()
    result = converter.pgm_input_data["three_winding_transformer"]

    # Assert
    assert result[0]["tap_side"] == Branch3Side.side_1.value
    assert result[1]["tap_side"] == Branch3Side.side_2.value
    assert result[2]["tap_side"] == Branch3Side.side_3.value
    assert result[3]["tap_side"] == Branch3Side.side_1.value
    assert result[4]["tap_side"] == Branch3Side.side_1.value
    assert result[5]["tap_side"] == Branch3Side.side_1.value
    assert result[0]["tap_pos"] == 34.0 != result[0]["tap_nom"]
    assert result[1]["tap_pos"] == 34.0 != result[1]["tap_nom"]
    assert result[2]["tap_pos"] == 34.0 != result[2]["tap_nom"]
    assert result[3]["tap_pos"] == 0 == result[3]["tap_nom"]
    assert result[4]["tap_pos"] == 0 == result[4]["tap_nom"]
    assert result[5]["tap_pos"] == 0 == result[5]["tap_nom"]
    assert result[6]["tap_size"] == 0

    # Default yndd for odd clocks
    assert result[7]["winding_1"] == WindingType.wye_n
    assert result[7]["winding_2"] == WindingType.delta
    assert result[7]["winding_3"] == WindingType.delta
    # Default ynynyn for even clocks
    assert result[8]["winding_1"] == WindingType.wye_n
    assert result[8]["winding_2"] == WindingType.wye_n
    assert result[8]["winding_3"] == WindingType.wye_n
    # Default ynynd for clock_12 even clock_13 odd
    assert result[9]["winding_1"] == WindingType.wye_n
    assert result[9]["winding_2"] == WindingType.wye_n
    assert result[9]["winding_3"] == WindingType.delta
    # Default yndyn for clock_12 odd clock_13 even
    assert result[10]["winding_1"] == WindingType.wye_n
    assert result[10]["winding_2"] == WindingType.delta
    assert result[10]["winding_3"] == WindingType.wye_n

    assert result[11]["clock_12"] == result[11]["clock_13"] == 2
    assert result[12]["clock_12"] == result[12]["clock_13"] == 1


@pytest.mark.parametrize(
    "kwargs",
    [
        {"vk0_hv_percent": 1},
        {"vkr0_hv_percent": 2},
        {"vk0_mv_percent": 3},
        {"vkr0_mv_percent": 4},
        {"vk0_lv_percent": 5},
        {"vkr0_lv_percent": 6},
    ],
)
@patch(
    "power_grid_model_io.converters.pandapower_converter.PandaPowerConverter.get_trafo3w_switch_states",
    new=MagicMock(return_value=pd.DataFrame({"side_1": [True], "side_2": [True], "side_3": [True]})),
)
@patch(
    "power_grid_model_io.converters.pandapower_converter.PandaPowerConverter.get_trafo3w_winding_types",
    new=MagicMock(return_value=pd.DataFrame({"winding_1": [0], "winding_2": [0], "winding_3": [0]})),
)
@patch(
    "power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._generate_ids",
    new=MagicMock(return_value=np.arange(1)),
)
@patch(
    "power_grid_model_io.converters.pandapower_converter.PandaPowerConverter._get_pgm_ids",
    new=MagicMock(return_value=pd.Series([0])),
)
def test_create_pgm_input_transformers3w__zero_sequence(kwargs) -> None:
    # Arrange
    pp_net: pp.pandapowerNet = pp.create_empty_network()
    pp.create_bus(net=pp_net, vn_kv=0.0)
    args = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    pp.create_transformer3w_from_parameters(pp_net, *args, **kwargs)

    converter = PandaPowerConverter()
    converter.pp_input_data = {k: v for k, v in pp_net.items() if isinstance(v, pd.DataFrame)}

    # Act
    with patch("power_grid_model_io.converters.pandapower_converter.logger") as mock_logger:
        converter._create_pgm_input_three_winding_transformers()
        mock_logger.warning.assert_called_once()


def test_create_pgm_input_three_winding_transformers__tap_at_star_point() -> None:
    # Arrange
    pp_net: pp.pandapowerNet = pp.create_empty_network()
    pp.create_bus(net=pp_net, vn_kv=0.0)
    args = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    pp.create_transformer3w_from_parameters(pp_net, *args, tap_at_star_point=True)

    converter = PandaPowerConverter()
    converter.pp_input_data = {k: v for k, v in pp_net.items() if isinstance(v, pd.DataFrame)}

    # Act/Assert
    with pytest.raises(RuntimeError, match="not supported"):
        converter._create_pgm_input_three_winding_transformers()


def test_create_pgm_input_three_winding_transformers__tap_dependent_impedance() -> None:
    # Arrange
    pp_net: pp.pandapowerNet = pp.create_empty_network()
    pp.create_bus(net=pp_net, vn_kv=0.0)
    args = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    pp.create_transformer3w_from_parameters(pp_net, *args, tap_dependent_impedance=True)

    converter = PandaPowerConverter()
    converter.pp_input_data = {k: v for k, v in pp_net.items() if isinstance(v, pd.DataFrame)}

    # Act/Assert
    with pytest.raises(RuntimeError, match="not supported"):
        converter._create_pgm_input_three_winding_transformers()


@patch("power_grid_model_io.converters.pandapower_converter.initialize_array")
def test_create_pgm_input_links(mock_init_array: MagicMock, converter):
    # Arrange
    # three switches, of which two switches (#1 and #3) are bus to bus switches
    pp_switches = pd.DataFrame(
        [[0, 0, "b", False], [0, 0, "l", False], [0, 0, "b", False]],
        index=[1, 2, 3],
        columns=["bus", "element", "et", "closed"],
    )
    converter.pp_input_data["switch"] = pp_switches

    # Act
    converter._create_pgm_input_links()

    # Assert

    # administration:
    converter._generate_ids.assert_called_once_with("switch", ANY, name="b2b_switches")
    pd.testing.assert_index_equal(converter._generate_ids.call_args_list[0].args[1], pd.Index([1, 3]))

    # initialization
    mock_init_array.assert_called_once_with(data_type="input", component_type="link", shape=2)

    # assignment:
    pgm: MagicMock = mock_init_array.return_value.__setitem__
    pgm.assert_any_call("id", ANY)
    pgm.assert_any_call("from_node", ANY)
    pgm.assert_any_call("to_node", ANY)
    pgm.assert_any_call("from_status", ANY)
    pgm.assert_any_call("to_node", ANY)
    assert len(pgm.call_args_list) == 5

    # result
    assert converter.pgm_input_data["link"] == mock_init_array.return_value


@patch("power_grid_model_io.converters.pandapower_converter.initialize_array")
def test_create_pgm_input_storage(mock_init_array: MagicMock, two_pp_objs, converter):
    # Arrange
    converter.pp_input_data["storage"] = two_pp_objs

    # Act / Assert
    with pytest.raises(NotImplementedError, match=r"Storage.*not implemented"):
        converter._create_pgm_input_storages()

    # initialization
    mock_init_array.assert_not_called()


@patch("power_grid_model_io.converters.pandapower_converter.initialize_array")
def test_create_pgm_input_impedance(mock_init_array: MagicMock, two_pp_objs, converter):
    # Arrange
    converter.pp_input_data["impedance"] = two_pp_objs

    # Act / Assert
    with pytest.raises(NotImplementedError, match=r"Impedance.*not implemented"):
        converter._create_pgm_input_impedances()

    # initialization
    mock_init_array.assert_not_called()


@patch("power_grid_model_io.converters.pandapower_converter.initialize_array")
def test_create_pgm_input_wards(mock_init_array: MagicMock, two_pp_objs, converter):
    # Arrange
    converter.pp_input_data["ward"] = two_pp_objs
    pgm_attr = ["id", "node", "status", "p_specified", "q_specified", "type"]
    pgm = {attr: MagicMock() for attr in pgm_attr}
    mock_init_array.return_value = pgm
    slices = [slice(None, 2), slice(-2, None)]
    assert slices[0].indices(2 * 2) == (0, 2, 1)
    assert slices[1].indices(2 * 2) == (2, 4, 1)

    # Act
    converter._create_pgm_input_wards()

    # Assert

    # administration:

    # initialization
    mock_init_array.assert_called_once_with(data_type="input", component_type="sym_load", shape=2 * 2)

    # retrieval:
    converter._get_pp_attr.assert_any_call("ward", "bus", expected_type="u4")
    converter._get_pp_attr.assert_any_call("ward", "ps_mw", expected_type="f8")
    converter._get_pp_attr.assert_any_call("ward", "qs_mvar", expected_type="f8")
    converter._get_pp_attr.assert_any_call("ward", "pz_mw", expected_type="f8")
    converter._get_pp_attr.assert_any_call("ward", "qz_mvar", expected_type="f8")
    converter._get_pp_attr.assert_any_call("ward", "in_service", expected_type="bool", default=True)
    assert len(converter._get_pp_attr.call_args_list) == 6

    # assignment:
    for attr in pgm_attr:
        for s in slices:
            pgm[attr].__setitem__.assert_any_call(s, ANY)
        assert len(pgm[attr].__setitem__.call_args_list) == len(slices)

    # result
    assert converter.pgm_input_data["sym_load"] == pgm


def test_create_pgm_input_wards__existing_loads() -> None:
    converter = PandaPowerConverter()
    # Arrange

    pp_net: pp.pandapowerNet = pp.create_empty_network()
    pp.create_bus(net=pp_net, vn_kv=0.0)
    pp.create_load(pp_net, 0, 0)
    pp.create_ward(pp_net, 0, 0, 0, 0, 0)

    converter.pp_input_data = pp_net

    # Act
    converter._create_pgm_input_nodes()
    converter._create_pgm_input_sym_loads()
    converter._create_pgm_input_wards()

    # assert
    assert len(converter.pgm_input_data["sym_load"]) == 5


@patch("power_grid_model_io.converters.pandapower_converter.initialize_array")
def test_create_pgm_input_xward(mock_init_array: MagicMock, two_pp_objs, converter):
    # Arrange
    converter.pp_input_data["xward"] = two_pp_objs

    # Act / Assert
    with pytest.raises(NotImplementedError, match=r"Extended Ward.*not implemented"):
        converter._create_pgm_input_xwards()

    # initialization
    mock_init_array.assert_not_called()


@patch("power_grid_model_io.converters.pandapower_converter.initialize_array")
def test_create_pgm_input_generators(mock_init_array: MagicMock, two_pp_objs, converter):
    # Arrange
    converter.pp_input_data["gen"] = two_pp_objs

    # Act / Assert
    with pytest.raises(NotImplementedError, match=r"Generators.*not implemented"):
        converter._create_pgm_input_generators()

    # initialization
    mock_init_array.assert_not_called()


@patch("power_grid_model_io.converters.pandapower_converter.initialize_array")
def test_create_pgm_input_dclines(mock_init_array: MagicMock, two_pp_objs, converter):
    # Arrange
    converter.pp_input_data["dcline"] = two_pp_objs

    # Act / Assert
    with pytest.raises(NotImplementedError, match=r"DC line .*not implemented"):
        converter._create_pgm_input_dclines()

    # initialization
    mock_init_array.assert_not_called()


@patch("power_grid_model_io.converters.pandapower_converter.initialize_array")
@patch("power_grid_model_io.converters.pandapower_converter.np.divide", new=lambda x, _, **kwargs: x)
@patch("power_grid_model_io.converters.pandapower_converter.np.power", new=lambda x, _, **kwargs: x)
@patch("power_grid_model_io.converters.pandapower_converter.np.sqrt", new=lambda x, **kwargs: x)
def test_create_pgm_input_motors(mock_init_array: MagicMock, two_pp_objs, converter):
    # Arrange
    converter.pp_input_data["motor"] = two_pp_objs

    # Act
    converter._create_pgm_input_motors()

    # Assert

    # administration:
    converter._generate_ids.assert_called_once_with("motor", two_pp_objs.index, name="motor_load")

    # initialization
    mock_init_array.assert_called_once_with(data_type="input", component_type="sym_load", shape=2)

    # retrieval:
    converter._get_pp_attr.assert_any_call("motor", "bus", expected_type="i8")
    converter._get_pp_attr.assert_any_call("motor", "pn_mech_mw", expected_type="f8")
    converter._get_pp_attr.assert_any_call("motor", "cos_phi", expected_type="f8")
    converter._get_pp_attr.assert_any_call("motor", "efficiency_percent", expected_type="f8")
    converter._get_pp_attr.assert_any_call("motor", "loading_percent", expected_type="f8")
    converter._get_pp_attr.assert_any_call("motor", "scaling", expected_type="f8")
    converter._get_pp_attr.assert_any_call("motor", "in_service", expected_type="bool", default=True)
    assert len(converter._get_pp_attr.call_args_list) == 7

    # assignment:
    pgm: MagicMock = mock_init_array.return_value.__setitem__
    pgm.assert_any_call("id", ANY)
    pgm.assert_any_call("node", ANY)
    pgm.assert_any_call("status", ANY)
    pgm.assert_any_call("p_specified", ANY)
    pgm.assert_any_call("q_specified", ANY)
    pgm.assert_any_call("type", ANY)
    assert len(pgm.call_args_list) == 6

    # result
    assert converter.pgm_input_data["sym_load"] == mock_init_array.return_value


def test_create_pgm_input_motors__existing_loads() -> None:
    converter = PandaPowerConverter()
    # Arrange

    pp_net: pp.pandapowerNet = pp.create_empty_network()
    pp.create_bus(net=pp_net, vn_kv=0.0)
    pp.create_load(pp_net, 0, 0)
    pp.create_motor(pp_net, 0, 0, 0)

    converter.pp_input_data = pp_net

    # Act
    converter._create_pgm_input_nodes()
    converter._create_pgm_input_sym_loads()
    converter._create_pgm_input_motors()

    # assert
    assert len(converter.pgm_input_data["sym_load"]) == 4


@pytest.mark.parametrize(
    "create_fn",
    [
        PandaPowerConverter._create_pgm_input_sources,
        PandaPowerConverter._create_pgm_input_shunts,
        PandaPowerConverter._create_pgm_input_lines,
        PandaPowerConverter._create_pgm_input_sym_gens,
        PandaPowerConverter._create_pgm_input_sym_loads,
        PandaPowerConverter._create_pgm_input_asym_gens,
        PandaPowerConverter._create_pgm_input_asym_loads,
        PandaPowerConverter._create_pgm_input_impedances,
        PandaPowerConverter._create_pgm_input_links,
        PandaPowerConverter._create_pgm_input_motors,
        PandaPowerConverter._create_pgm_input_nodes,
        PandaPowerConverter._create_pgm_input_storages,
        PandaPowerConverter._create_pgm_input_three_winding_transformers,
        PandaPowerConverter._create_pgm_input_transformers,
        PandaPowerConverter._create_pgm_input_wards,
        PandaPowerConverter._create_pgm_input_xwards,
        PandaPowerConverter._create_pgm_input_sources,
        PandaPowerConverter._create_pgm_input_generators,
        PandaPowerConverter._create_pgm_input_dclines,
    ],
)
def test_create_pp_input_object__empty(create_fn: Callable[[PandaPowerConverter], None]):
    # Arrange: No table
    converter = PandaPowerConverter()
    converter.pp_input_data = pp.create_empty_network()

    # Act / Assert
    with patch("power_grid_model_io.converters.pandapower_converter.initialize_array") as mock_init_array:
        create_fn(converter)
        mock_init_array.assert_not_called()


def test_generate_ids():
    converter = PandaPowerConverter()
    test_table = pd.DataFrame(
        [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
        columns=["col1", "col2", "col3"],
        index=[11, 12, 13],
    )
    converter.pp_input_data["test_table"] = test_table
    converter.next_idx = 1
    pgm_idx_actual = converter._generate_ids("test_table", test_table.index, name="ids_name")
    pgm_idx_expected = np.array([1, 2, 3], dtype=np.int32)

    assert converter.next_idx == 4
    pd.testing.assert_series_equal(
        converter.idx[("test_table", "ids_name")], pd.Series(pgm_idx_expected, index=test_table.index)
    )
    pd.testing.assert_series_equal(
        converter.idx_lookup[("test_table", "ids_name")], pd.Series(test_table.index, index=pgm_idx_expected)
    )
    np.testing.assert_array_equal(pgm_idx_actual, pgm_idx_expected)


def test_get_pgm_ids():
    # Arrange
    converter = PandaPowerConverter()
    converter.idx = {
        ("bus", None): pd.Series([10, 11, 12], index=[0, 1, 2]),
        ("load", "const_current"): pd.Series([13, 14], index=[3, 4]),
    }

    # Act
    bus_ids = converter._get_pgm_ids(pp_table="bus", pp_idx=pd.Series([2, 1]))
    load_ids = converter._get_pgm_ids(pp_table="load", name="const_current", pp_idx=pd.Series([3]))
    all_bus_ids = converter._get_pgm_ids(pp_table="bus")

    # Assert
    pd.testing.assert_series_equal(bus_ids, pd.Series([12, 11], index=[2, 1]))
    pd.testing.assert_series_equal(load_ids, pd.Series([13], index=[3]))
    pd.testing.assert_series_equal(all_bus_ids, pd.Series([10, 11, 12], index=[0, 1, 2]))


def test_get_pgm_ids__key_error():
    # Arrange
    converter = PandaPowerConverter()

    # Act / Assert
    with pytest.raises(KeyError, match=r"index.*bus"):
        converter._get_pgm_ids(pp_table="bus")


def test_get_pp_ids():
    # Arrange
    converter = PandaPowerConverter()
    converter.idx_lookup = {
        ("bus", None): pd.Series([0, 1, 2], index=[10, 11, 12]),
        ("load", "const_current"): pd.Series([3, 4], index=[13, 14]),
    }

    # Act
    bus_ids = converter._get_pp_ids(pp_table="bus", pgm_idx=pd.Series([12, 11]))
    load_ids = converter._get_pp_ids(pp_table="load", name="const_current", pgm_idx=pd.Series([13]))
    all_bus_ids = converter._get_pp_ids(pp_table="bus")

    # Assert
    pd.testing.assert_series_equal(bus_ids, pd.Series([2, 1], index=[12, 11]))
    pd.testing.assert_series_equal(load_ids, pd.Series([3], index=[13]))
    pd.testing.assert_series_equal(all_bus_ids, pd.Series([0, 1, 2], index=[10, 11, 12]))


def test_get_pp_ids__key_error():
    # Arrange
    converter = PandaPowerConverter()

    # Act / Assert
    with pytest.raises(KeyError, match=r"index.*bus"):
        converter._get_pp_ids(pp_table="bus")


def test_get_tap_size():
    # Arrange
    pp_trafo = pd.DataFrame(
        [["hv", 5.0, 10.5, 0.4], ["lv", 10.0, 10.5, 0.4], ["lv", np.nan, 10.5, 0.4]],
        columns=["tap_side", "tap_step_percent", "vn_hv_kv", "vn_lv_kv"],
    )
    expected_tap_size = np.array([525.0, 40.0, np.nan], dtype=np.float64)

    # Act
    actual_tap_size = PandaPowerConverter._get_tap_size(pp_trafo)

    # Assert
    np.testing.assert_array_equal(actual_tap_size, expected_tap_size)


def test_get_transformer_tap_side():
    # Arrange
    pp_trafo_tap_side = np.array(["hv", "lv", "lv", "hv"])
    expected_tap_side = np.array([0, 1, 1, 0], dtype=np.int8)

    # Act
    actual_tap_side = PandaPowerConverter._get_transformer_tap_side(pp_trafo_tap_side)

    # Assert
    np.testing.assert_array_equal(actual_tap_side, expected_tap_side)


def test_get_3wtransformer_tap_side():
    # Arrange
    pp_trafo3w_tap_side = np.array(["hv", "mv", "lv", None, "mv", "lv", "hv", "lv", None])
    expected_tap_side = np.array([0, 1, 2, 0, 1, 2, 0, 2, 0], dtype=np.int8)

    # Act
    actual_tap_side = PandaPowerConverter._get_3wtransformer_tap_side(pp_trafo3w_tap_side)

    # Assert
    np.testing.assert_array_equal(actual_tap_side, expected_tap_side)


def test_get_3wtransformer_tap_size():
    # Arrange
    pp_trafo = pd.DataFrame(
        [["hv", 62.0, 10.5, 400.0, 32.1], ["lv", 62.0, 10.5, 400.0, 32.1], ["mv", 62.0, 10.5, 400.0, 32.1]],
        columns=["tap_side", "tap_step_percent", "vn_hv_kv", "vn_mv_kv", "vn_lv_kv"],
    )
    expected_tap_size = np.array([6510.0, 19902.0, 248000.0], dtype=np.float64)

    # Act
    actual_tap_size = PandaPowerConverter._get_3wtransformer_tap_size(pp_trafo)

    # Assert
    np.testing.assert_array_equal(actual_tap_size, expected_tap_size)


@patch("power_grid_model_io.converters.pandapower_converter.get_winding")
def test_get_trafo_winding_types__vector_group(mock_get_winding: MagicMock):
    # Arrange
    converter = PandaPowerConverter()
    converter.pp_input_data = {
        "trafo": pd.DataFrame([(1, "Dyn"), (2, "YNd"), (3, "Dyn")], columns=["id", "vector_group"])
    }
    mock_get_winding.side_effect = [WindingType.delta, WindingType.wye_n, WindingType.wye_n, WindingType.delta]
    expected = pd.DataFrame([(2, 1), (1, 2), (2, 1)], columns=["winding_from", "winding_to"])

    # Act
    actual = converter.get_trafo_winding_types()

    # Assert
    pd.testing.assert_frame_equal(actual, expected)
    assert len(mock_get_winding.call_args_list) == 4
    assert mock_get_winding.call_args_list[0] == call("D")
    assert mock_get_winding.call_args_list[1] == call("yn")
    assert mock_get_winding.call_args_list[2] == call("YN")
    assert mock_get_winding.call_args_list[3] == call("d")


def test_get_trafo_winding_types__vector_group_missing():
    # Arrange
    converter = PandaPowerConverter()
    converter.pp_input_data = {"trafo": pd.DataFrame([1, 2, 3], columns=["id"])}
    expected = pd.DataFrame(np.full((3, 2), np.nan), columns=["winding_from", "winding_to"])

    # Act
    actual = converter.get_trafo_winding_types()

    # Assert
    pd.testing.assert_frame_equal(actual, expected)


@patch("power_grid_model_io.converters.pandapower_converter.get_winding")
def test_get_trafo3w_winding_types__vector_group(mock_get_winding: MagicMock):
    # Arrange
    converter = PandaPowerConverter()
    converter.pp_input_data = {
        "trafo3w": pd.DataFrame([(1, "Dynz"), (2, "YNdy"), (3, "Dyny")], columns=["id", "vector_group"])
    }
    mock_get_winding.side_effect = [
        WindingType.delta,
        WindingType.wye_n,
        WindingType.zigzag,
        WindingType.wye_n,
        WindingType.delta,
        WindingType.wye,
        WindingType.delta,
        WindingType.wye_n,
        WindingType.wye,
    ]

    expected = pd.DataFrame([[2, 1, 3], [1, 2, 0], [2, 1, 0]], columns=["winding_1", "winding_2", "winding_3"])

    # Act
    actual = converter.get_trafo3w_winding_types()

    # Assert
    pd.testing.assert_frame_equal(actual, expected)
    assert len(mock_get_winding.call_args_list) == 9
    assert mock_get_winding.call_args_list[0] == call("D")
    assert mock_get_winding.call_args_list[1] == call("yn")
    assert mock_get_winding.call_args_list[2] == call("z")
    assert mock_get_winding.call_args_list[3] == call("YN")
    assert mock_get_winding.call_args_list[4] == call("d")
    assert mock_get_winding.call_args_list[5] == call("y")
    assert mock_get_winding.call_args_list[6] == call("D")
    assert mock_get_winding.call_args_list[7] == call("yn")
    assert mock_get_winding.call_args_list[8] == call("y")


def test_get_trafo3w_winding_types__vector_group_missing():
    # Arrange
    converter = PandaPowerConverter()
    converter.pp_input_data = {"trafo3w": pd.DataFrame([1, 2, 3], columns=["id"])}
    expected = pd.DataFrame(np.full((3, 3), np.nan), columns=["winding_1", "winding_2", "winding_3"])

    # Act
    actual = converter.get_trafo3w_winding_types()

    # Assert
    pd.testing.assert_frame_equal(actual, expected)


def test_get_winding_types__value_error():
    # Arrange
    converter = PandaPowerConverter()
    converter.pp_input_data = {"trafo": pd.DataFrame([(1, "ADyn")], columns=["id", "vector_group"])}

    # Act / Assert
    with pytest.raises(ValueError):
        converter.get_trafo_winding_types()


def test_get_trafo3w_winding_types__value_error():
    # Arrange
    converter = PandaPowerConverter()
    converter.pp_input_data = {"trafo3w": pd.DataFrame([(1, "ADyndrr")], columns=["id", "vector_group"])}

    # Act / Assert
    with pytest.raises(ValueError):
        converter.get_trafo3w_winding_types()


def test_get_individual_switch_states():
    # Arrange
    pp_trafo = pd.DataFrame(
        columns=["index", "hv_bus"],
        data=[[1, 101], [2, 102], [3, 103]],
    )
    pp_switches = pd.DataFrame(
        columns=["element", "bus", "closed"],
        data=[[1, 101, False], [2, 202, False], [3, 103, True]],
    )

    expected_state = pd.Series([False, True, True], dtype=bool)

    # Act
    actual_state = PandaPowerConverter.get_individual_switch_states(pp_trafo, pp_switches, "hv_bus")

    # Assert
    np.testing.assert_array_equal(actual_state, expected_state)


def test_get_id():
    # Arrange
    converter = PandaPowerConverter()
    converter.idx = {("line", None): pd.Series([21, 345, 0, 3, 15], index=[0, 1, 2, 3, 4])}

    # Act
    actual_id = converter.get_id("line", 1)

    # Assert
    np.testing.assert_array_equal(actual_id, 345)


@patch("power_grid_model_io.converters.pandapower_converter.PandaPowerConverter.get_individual_switch_states")
def test_get_switch_states_lines(mock_get_individual_switch_states: MagicMock):
    # Arrange
    converter = PandaPowerConverter()
    converter.pp_input_data = {
        "line": pd.DataFrame(columns=["from_bus", "to_bus"], data=[[101, 102]], index=[1]),
        "switch": pd.DataFrame(
            columns=["bus", "et", "element", "closed"],
            data=[[101, "l", 1, False], [102, "x", 1, False]],
            index=[1001, 1002],
        ),
    }
    mock_get_individual_switch_states.side_effect = [
        pd.Series([False], index=[1], dtype=bool, name="closed"),
        pd.Series([True], index=[1], dtype=bool, name="closed"),
    ]
    expected = pd.DataFrame(columns=["from", "to"], index=[1], data=[[False, True]])

    # Act
    actual = converter.get_switch_states("line")

    # Assert
    pd.testing.assert_frame_equal(actual, expected)

    assert len(mock_get_individual_switch_states.call_args_list) == 2

    assert mock_get_individual_switch_states.call_args_list[0] == call(ANY, ANY, "from_bus")
    pd.testing.assert_frame_equal(
        mock_get_individual_switch_states.call_args_list[0].args[0],
        pd.DataFrame(columns=["index", "from_bus"], data=[[1, 101]], index=[1]),
    )
    pd.testing.assert_frame_equal(
        mock_get_individual_switch_states.call_args_list[0].args[1],
        pd.DataFrame(columns=["element", "bus", "closed"], data=[[1, 101, False]], index=[1001]),
    )

    assert mock_get_individual_switch_states.call_args_list[1] == call(ANY, ANY, "to_bus")
    pd.testing.assert_frame_equal(
        mock_get_individual_switch_states.call_args_list[1].args[0],
        pd.DataFrame(columns=["index", "to_bus"], data=[[1, 102]], index=[1]),
    )
    pd.testing.assert_frame_equal(
        mock_get_individual_switch_states.call_args_list[1].args[1],
        pd.DataFrame(columns=["element", "bus", "closed"], data=[[1, 101, False]], index=[1001]),
    )


@patch("power_grid_model_io.converters.pandapower_converter.PandaPowerConverter.get_individual_switch_states")
def test_get_switch_states_trafos(mock_get_individual_switch_states: MagicMock):
    # Arrange
    converter = PandaPowerConverter()
    converter.pp_input_data = {
        "trafo": pd.DataFrame([[2, 32, 31]], columns=["index", "hv_bus", "lv_bus"]),
        "switch": pd.DataFrame(
            [[101, 32, "t", 2, True], [321, 31, "t", 2, False]],
            columns=["index", "bus", "et", "element", "closed"],
        ),
    }
    mock_get_individual_switch_states.side_effect = [
        pd.Series([False], index=[2], dtype=bool, name="closed"),
        pd.Series([True], index=[2], dtype=bool, name="closed"),
    ]

    expected = pd.DataFrame(columns=["from", "to"], index=[2], data=[[False, True]])

    # Act
    actual = converter.get_switch_states("trafo")

    # Assert
    pd.testing.assert_frame_equal(actual, expected)
    assert len(mock_get_individual_switch_states.call_args_list) == 2


def test_get_switch_states__exception():
    # Arrange
    converter = PandaPowerConverter()

    # Act / Assert
    with pytest.raises(KeyError, match=r"link"):
        converter.get_switch_states("link")


@patch("power_grid_model_io.converters.pandapower_converter.PandaPowerConverter.get_individual_switch_states")
def test_get_trafo3w_switch_states(mock_get_individual_switch_states: MagicMock):
    # Arrange
    converter = PandaPowerConverter()
    converter.pp_input_data = {
        "trafo3w": pd.DataFrame([[2, 32, 31, 315]], columns=["index", "hv_bus", "mv_bus", "lv_bus"]),
        "switch": pd.DataFrame(
            [[101, 315, "t3", 2, False], [321, 32, "t3", 2, False]],
            columns=["index", "bus", "et", "element", "closed"],
        ),
    }
    mock_get_individual_switch_states.side_effect = [False, True, False]

    expected = pd.DataFrame(columns=["side_1", "side_2", "side_3"], data=[[False, True, False]])

    # Act
    actual = converter.get_trafo3w_switch_states(converter.pp_input_data["trafo3w"])

    # Assert
    pd.testing.assert_frame_equal(actual, expected)
    assert len(mock_get_individual_switch_states.call_args_list) == 3


def test_lookup_id():
    # Arrange
    converter = PandaPowerConverter()
    converter.idx_lookup = {
        ("line", None): pd.Series([0, 1, 2, 3, 4], index=[21, 345, 0, 3, 15]),
        ("load", "const_current"): pd.Series([5, 6, 7, 8, 9], index=[543, 14, 34, 48, 4]),
    }

    expected_line = {"table": "line", "index": 4}
    expected_load = {"table": "load", "name": "const_current", "index": 8}

    # Act
    actual_line = converter.lookup_id(15)
    actual_load = converter.lookup_id(48)

    # Assert
    np.testing.assert_array_equal(actual_line, expected_line)
    np.testing.assert_array_equal(actual_load, expected_load)


def test_lookup_id__value_error():
    # Arrange
    converter = PandaPowerConverter()
    converter.idx_lookup = {("line", None): pd.Series([0, 1, 2, 3, 4], index=[21, 345, 0, 3, 15])}

    # Act / Assert
    with pytest.raises(KeyError):
        converter.lookup_id(5)


def test_get_pp_attr_attribute_exists():
    # Arrange
    converter = PandaPowerConverter()
    converter.pp_input_data = {
        "trafo3w": pd.DataFrame([[2, 32, 31, 315]], columns=["index", "hv_bus", "mv_bus", "lv_bus"])
    }
    expected = np.array(32)

    # Act
    actual = converter._get_pp_attr("trafo3w", "hv_bus", expected_type="u4")

    # Assert
    np.testing.assert_array_equal(actual, expected)


def test_get_pp_attr_attribute_doesnt_exist():
    # Arrange
    converter = PandaPowerConverter()
    converter.pp_input_data = {"trafo3w": pd.DataFrame([[2, 31, 315]], columns=["index", "mv_bus", "lv_bus"])}

    # Act / Assert
    with pytest.raises(KeyError):
        converter._get_pp_attr("trafo3w", "hv_bus", expected_type="u4")


def test_get_pp_attr_use_default():
    # Arrange
    converter = PandaPowerConverter()
    converter.pp_input_data = {"trafo3w": pd.DataFrame([[2, 31, 315]], columns=["index", "mv_bus", "lv_bus"])}
    expected = np.array(625)

    # Act
    actual = converter._get_pp_attr("trafo3w", "hv_bus", expected_type="u4", default=625)

    # Assert
    np.testing.assert_array_equal(actual, expected)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0
"""
Common data types used in the Power Grid Model project
"""

from typing import Any, Dict, List, Union

ExtraInfo = Dict[int, Any]
"""
ExtraInfo is information about power grid model objects that are not part of the calculations. E.g. the original ID or
name of a node, or the material of a cable (line) etc.

It is a dictionary with numerical keys corresponding to the ids in input_data etc. The values are dictionaries with
textual keys. Their values may be anything, but it is advised to use only JSON serializable types like numerical values,
strings, lists, dictionaries etc.

    {
        1: {
            "length_km": 123.4,
            "material": "Aluminuminuminum",
        },
        2: {
            "id_reference": {
                "table": "load",
                "name": "const_power",
                "index": 101
            }
        }
    }
"""

ExtraInfoLookup = ExtraInfo
"""
Legacy type name; use ExtraInfo instead!
"""

StructuredData = Union[Dict[str, List[Dict[str, Any]]], List[Dict[str, List[Dict[str, Any]]]]]
"""
Structured data is a multi dimensional structure (component_type -> objects -> attribute -> value) or a list of those
dictionaries:
    {
      "node":
        [
          {"id": 0, "u_rated": 110000.0},
          {"id": 1, "u_rated": 110000.0},
        ],
      "line":
        [
          {"id": 2, "from_node": 0, "to_node": 1, "from_status": 1, "to_status": 1}
        ]
      "source":
        [
          {"id": 3, "node": 0, "status": 1, "u_ref": 1.0}
        ]
    }
"""



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0
from pathlib import Path
from unittest.mock import patch

import pytest

from power_grid_model_io.converters.vision_excel_converter import DEFAULT_MAPPING_FILE, VisionExcelConverter
from power_grid_model_io.utils.excel_ambiguity_checker import ExcelAmbiguityChecker


@pytest.fixture
def converter() -> VisionExcelConverter:
    # Arrange
    converter = VisionExcelConverter()
    converter._get_id("Nodes", {"Number": 1}, None)  # node: 0
    converter._get_id("Cables", {"Number": 1}, None)  # branch: 1
    converter._get_id("Links", {"Number": 1}, None)  # branch: 2
    converter._get_id("Reactance coils", {"Number": 1}, None)  # branch: 3
    converter._get_id("Special transformers", {"Number": 1}, None)  # branch: 4
    converter._get_id("Transformer loads", {"Node.Number": 1, "Subnumber": 2}, "transformer")  # virtual: 5
    converter._get_id("Transformer loads", {"Node.Number": 1, "Subnumber": 2}, "internal_node")  # virtual:  6
    converter._get_id("Transformer loads", {"Node.Number": 1, "Subnumber": 2}, "load")  # virtual:  7
    converter._get_id("Transformer loads", {"Node.Number": 1, "Subnumber": 2}, "generation")  # virtual:  8
    converter._get_id("Transformer loads", {"Node.Number": 1, "Subnumber": 2}, "pv_generation")  # virtual:  9
    converter._get_id("Sources", {"Node.Number": 1, "Subnumber": 2}, None)  # appliance: 10
    converter._get_id("Synchronous generators", {"Node.Number": 1, "Subnumber": 2}, None)  # appliance: 11
    converter._get_id("Wind turbines", {"Node.Number": 1, "Subnumber": 2}, None)  # appliance: 12
    converter._get_id("Loads", {"Node.Number": 1, "Subnumber": 2}, None)  # appliance: 13
    converter._get_id("Zigzag transformers", {"Node.Number": 1, "Subnumber": 2}, None)  # appliance: 14
    converter._get_id("Pvs", {"Node.Number": 1, "Subnumber": 2}, None)  # appliance: 15
    return converter


@pytest.mark.parametrize("language", ["en", "nl"])
def test_mapping_files_exist(language: str):
    vf = Path(str(DEFAULT_MAPPING_FILE).format(language=language))
    assert vf.exists()


def test_initialization():
    with pytest.raises(FileNotFoundError, match="No Vision Excel mapping available for language 'abcde'"):
        VisionExcelConverter(language="abcde")

    with patch(
        "power_grid_model_io.converters.tabular_converter.TabularConverter.set_mapping_file"
    ) as mock_set_mapping_file:
        VisionExcelConverter()
        mock_set_mapping_file.assert_called_once()

    with patch("power_grid_model_io.converters.vision_excel_converter.VisionExcelFileStore") as MockFileStore:
        VisionExcelConverter()
        MockFileStore.assert_not_called()

        VisionExcelConverter(source_file="source_file")
        MockFileStore.assert_called_once()


def test_id_lookup_exceptions():
    # Arrange
    converter = VisionExcelConverter()
    converter._id_reference = None

    # Act / Assert
    with pytest.raises(ValueError, match=r"Missing .* for VisionExcelConverter\.get_node_id\(\)"):
        converter.get_node_id(number=0)

    with pytest.raises(ValueError, match=r"Missing .* VisionExcelConverter\.get_branch_id\(\)"):
        converter.get_branch_id(table="", number=0)

    with pytest.raises(ValueError, match=r"Missing .* VisionExcelConverter\.get_appliance_id\(\)"):
        converter.get_appliance_id(table="", node_number=0, sub_number=0)

    with pytest.raises(ValueError, match=r"Missing .* VisionExcelConverter\.get_virtual_id\(\)"):
        converter.get_virtual_id(table="", obj_name="", node_number=0, sub_number=0)


def test_get_node_id(converter: VisionExcelConverter):
    # Act / Assert
    assert converter.get_node_id(number=1) == 0

    with pytest.raises(KeyError):
        converter.get_node_id(number=2)


def test_get_branch_id(converter: VisionExcelConverter):
    # Act / Assert
    assert converter.get_branch_id(table="Cables", number=1) == 1
    assert converter.get_branch_id(table="Links", number=1) == 2
    assert converter.get_branch_id(table="Reactance coils", number=1) == 3
    assert converter.get_branch_id(table="Special transformers", number=1) == 4

    with pytest.raises(KeyError):
        converter.get_branch_id(table="Dummy", number=1)

    with pytest.raises(KeyError):
        converter.get_branch_id(table="Cables", number=2)


def test_get_virtual_id(converter: VisionExcelConverter):
    # Act / Assert
    assert converter.get_virtual_id(table="Transformer loads", obj_name="transformer", node_number=1, sub_number=2) == 5
    assert (
        converter.get_virtual_id(table="Transformer loads", obj_name="internal_node", node_number=1, sub_number=2) == 6
    )
    assert converter.get_virtual_id(table="Transformer loads", obj_name="load", node_number=1, sub_number=2) == 7
    assert converter.get_virtual_id(table="Transformer loads", obj_name="generation", node_number=1, sub_number=2) == 8
    assert (
        converter.get_virtual_id(table="Transformer loads", obj_name="pv_generation", node_number=1, sub_number=2) == 9
    )

    with pytest.raises(KeyError):
        converter.get_virtual_id(table="Dummy", obj_name="transformer", node_number=1, sub_number=2)

    with pytest.raises(KeyError):
        converter.get_virtual_id(table="Transformer loads", obj_name="Dummy", node_number=1, sub_number=2)

    with pytest.raises(KeyError):
        converter.get_virtual_id(table="Transformer loads", obj_name="transformer", node_number=2, sub_number=2)

    with pytest.raises(KeyError):
        converter.get_virtual_id(table="Transformer loads", obj_name="transformer", node_number=1, sub_number=3)


def test_get_appliance_id(converter: VisionExcelConverter):
    # Act / Assert
    assert converter.get_appliance_id(table="Sources", node_number=1, sub_number=2) == 10
    assert converter.get_appliance_id(table="Synchronous generators", node_number=1, sub_number=2) == 11
    assert converter.get_appliance_id(table="Wind turbines", node_number=1, sub_number=2) == 12
    assert converter.get_appliance_id(table="Loads", node_number=1, sub_number=2) == 13
    assert converter.get_appliance_id(table="Zigzag transformers", node_number=1, sub_number=2) == 14
    assert converter.get_appliance_id(table="Pvs", node_number=1, sub_number=2) == 15

    with pytest.raises(KeyError):
        converter.get_appliance_id(table="Dummy", node_number=1, sub_number=2)

    with pytest.raises(KeyError):
        converter.get_appliance_id(table="Sources", node_number=2, sub_number=2)

    with pytest.raises(KeyError):
        converter.get_appliance_id(table="Sources", node_number=1, sub_number=3)


def test_ambiguity_in_vision_excel():
    ambiguious_test_file = Path(__file__).parents[2] / "data" / "vision" / "excel_ambiguity_check_data.xlsx"
    excel_file_checker = ExcelAmbiguityChecker(file_path=ambiguious_test_file.as_posix())
    res, _ = excel_file_checker.check_ambiguity()
    assert res == True



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0
"""
Excel File Store
"""

import re
from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple, Union

import pandas as pd

from power_grid_model_io.data_stores.base_data_store import (
    DICT_KEY_NUMBER,
    DICT_KEY_SUBNUMBER,
    VISION_EXCEL_LAN_DICT,
    BaseDataStore,
)
from power_grid_model_io.data_types import LazyDataFrame, TabularData
from power_grid_model_io.utils.uuid_excel_cvtr import (
    UUID2IntCvtr,
    add_guid_values_to_cvtr,
    get_special_key_map,
    special_nodes_en,
    special_nodes_nl,
    update_column_names,
)


class ExcelFileStore(BaseDataStore[TabularData]):
    """
    Excel File Store

    The first row of each sheet is expected to contain the column names, unless specified differently by an extension
    of this class. Columns with duplicate names (on the same sheet) are either removed (if they contain exactly the
    same values) or renamed.
    """

    __slots__ = ("_file_paths", "_excel_files", "_header_rows")

    _unnamed_pattern: re.Pattern = re.compile(r"Unnamed: \d+_level_\d+")

    def __init__(
        self,
        file_path: Optional[Path] = None,
        *,
        language: str = "en",
        terms_changed: Optional[dict] = None,
        **extra_paths: Path,
    ):
        super().__init__()
        # Create a dictionary of all supplied file paths:
        # {"": file_path, extra_name[0]: extra_path[0], extra_name[1]: extra_path[1], ...}
        self._file_paths: Dict[str, Path] = {}
        if file_path is not None:
            self._file_paths[""] = file_path
        for name, path in extra_paths.items():
            self._file_paths[name] = path

        for name, path in self._file_paths.items():
            if path.suffix.lower() not in {".xls", ".xlsx"}:
                name = name.title() if name else "Excel"
                raise ValueError(f"{name} file should be a .xls or .xlsx file, {path.suffix} provided.")

        self._header_rows: List[int] = [0]
        self._language = language
        self._vision_excel_key_mapping = VISION_EXCEL_LAN_DICT[self._language]
        self._terms_changed = terms_changed if terms_changed is not None else {}
        self._uuid_cvtr = UUID2IntCvtr()

    def files(self) -> Dict[str, Path]:
        """
        The files as supplied in the constructor. Note that the file names are read-only.

        Returns: A copy of the file paths as set in the constructor.
        """
        return self._file_paths.copy()

    def load(self) -> TabularData:
        """
        Load one or more Excel file as tabular data.

        Returns: The contents of all the Excel file supplied in the constructor. The tables of the main file will
        have no prefix, while the tables of all the extra files will be prefixed with the name of the key word argument
        as supplied in the constructor.
        """

        def lazy_sheet_loader(xls_file: pd.ExcelFile, xls_sheet_name: str):
            def sheet_loader():
                sheet_data = xls_file.parse(xls_sheet_name, header=self._header_rows)
                sheet_data = self._remove_unnamed_column_placeholders(data=sheet_data)
                sheet_data = self._handle_duplicate_columns(data=sheet_data, sheet_name=xls_sheet_name)
                sheet_data = self._process_uuid_columns(data=sheet_data, sheet_name=xls_sheet_name)
                sheet_data = self._update_column_names(data=sheet_data)
                return sheet_data

            return sheet_loader

        data: Dict[str, LazyDataFrame] = {}
        for name, path in self._file_paths.items():
            excel_file = pd.ExcelFile(path)
            for sheet_name in excel_file.sheet_names:
                loader = lazy_sheet_loader(excel_file, sheet_name)
                if name != "":  # If the Excel file is not the main file, prefix the sheet name with the file name
                    sheet_name = f"{name}.{sheet_name}"
                if sheet_name in data:
                    raise ValueError(f"Duplicate sheet name '{sheet_name}'")
                data[sheet_name] = loader
        return TabularData(**data)

    def save(self, data: TabularData) -> None:
        """
        Store tabular data as one or more Excel file.

        Args:
            data: Tha data to store. The keys of the tables will be the names of the spread sheets in the excel
            files. Table names with a prefix corresponding to the name of the key word argument as supplied in the
            constructor will be stored in the associated files.
        """

        # First group all sheets per file. Each sheet name that starts with a file name is assigned to that file.
        # Sheets that don't start with a file name are assigned to the first file.
        sheets: Dict[str, Dict[str, pd.DataFrame]] = {file_name: {} for file_name in self._file_paths}
        for sheet_name, sheet_data in data.items():
            if "." in sheet_name:
                file_name, alt_sheet_name = sheet_name.split(".", maxsplit=1)
                if file_name in self._file_paths:
                    sheets[file_name][alt_sheet_name] = sheet_data
                    continue
            sheets[""][sheet_name] = sheet_data

        # Create an Excel file if there is at least one sheet.
        for file_name, file_path in self._file_paths.items():
            if not sheets[file_name]:
                continue
            with pd.ExcelWriter(path=file_path) as excel_writer:  # pylint: disable=abstract-class-instantiated
                for sheet_name, sheet_data in sheets[file_name].items():
                    if not isinstance(sheet_data, pd.DataFrame):
                        sheet_data = pd.DataFrame(sheet_data)
                    sheet_data.to_excel(
                        excel_writer=excel_writer,
                        sheet_name=sheet_name,
                    )

    def _remove_unnamed_column_placeholders(self, data: pd.DataFrame) -> pd.DataFrame:
        if data.empty:
            return data

        def is_unnamed(col_name) -> bool:
            return bool(self._unnamed_pattern.fullmatch(str(col_name)))

        if data.columns.nlevels == 1:
            data.columns = pd.Index("" if is_unnamed(idx) else idx for idx in data.columns.values)
        else:
            data.columns = pd.MultiIndex.from_tuples(
                [tuple("" if is_unnamed(idx) else idx for idx in col_idx) for col_idx in data.columns.values]
            )

        return data

    def _handle_duplicate_columns(self, data: pd.DataFrame, sheet_name: str) -> pd.DataFrame:
        if data.empty:
            return data

        to_rename = self._check_duplicate_values(sheet_name=sheet_name, data=data)
        if to_rename:
            columns = data.columns.values.copy()
            for col_idx, new_name in to_rename.items():
                new_name = new_name[0] if isinstance(new_name, tuple) else new_name
                full_new_name = (new_name, columns[col_idx][1])
                self._log.warning(
                    "Column is renamed",
                    sheet_name=sheet_name,
                    col_name=columns[col_idx],
                    new_name=full_new_name,
                    col_idx=col_idx,
                )
                columns[col_idx] = full_new_name

            if data.columns.nlevels == 1:
                data.columns = pd.Index(columns)
            else:
                data.columns = pd.MultiIndex.from_tuples(columns)

        return data

    def _check_duplicate_values(self, sheet_name: str, data: pd.DataFrame) -> Dict[int, Union[str, Tuple[str, ...]]]:
        grouped = self._group_columns_by_index(data=data)

        to_rename: Dict[int, Union[str, Tuple[str, ...]]] = {}

        for col_name, col_idxs in grouped.items():
            # No duplicate column names
            if len(col_idxs) == 1:
                continue

            # Select the first column as a reference
            ref_idx = min(col_idxs)

            # Select the rest as duplicates
            dup_idxs = col_idxs - {ref_idx}

            same_values = all(data.iloc[:, dup_idx].equals(data.iloc[:, ref_idx]) for dup_idx in dup_idxs)
            if same_values:
                self._log.warning(
                    "Found duplicate column names, with same data",
                    sheet_name=sheet_name,
                    col_name=col_name,
                    col_idx=sorted(col_idxs),
                )
            else:
                self._log.error(
                    "Found duplicate column names, with different data",
                    sheet_name=sheet_name,
                    col_name=col_name,
                    col_idx=sorted(col_idxs),
                )
            for counter, dup_idx in enumerate(sorted(dup_idxs), start=2):
                if isinstance(col_name, tuple):
                    to_rename[dup_idx] = (f"{col_name[0]}_{counter}",) + col_name[1:]
                else:
                    to_rename[dup_idx] = f"{col_name}_{counter}"

        return to_rename

    def _process_uuid_columns(self, data: pd.DataFrame, sheet_name: str) -> pd.DataFrame:
        first_level = data.columns.get_level_values(0)
        guid_columns = first_level[first_level.str.endswith("GUID")]

        sheet_key_mapping = get_special_key_map(
            sheet_name=sheet_name, nodes_en=special_nodes_en, nodes_nl=special_nodes_nl
        )

        for guid_column in guid_columns:
            nr = VISION_EXCEL_LAN_DICT[self._language][DICT_KEY_NUMBER]
            add_guid_values_to_cvtr(data, guid_column, self._uuid_cvtr)
            new_column_name = guid_column.replace("GUID", nr)
            if guid_column == "GUID" and sheet_key_mapping not in (None, {}):
                new_column_name = guid_column.replace("GUID", sheet_key_mapping[DICT_KEY_SUBNUMBER])
            guid_column_pos = first_level.tolist().index(guid_column)
            try:
                data.insert(guid_column_pos + 1, new_column_name, data[guid_column].apply(self._uuid_cvtr.query))
            except ValueError:
                data[new_column_name] = data[guid_column].apply(self._uuid_cvtr.query)

        return data

    def _update_column_names(self, data: pd.DataFrame) -> pd.DataFrame:
        update_column_names(data, self._terms_changed)
        return data

    @staticmethod
    def _group_columns_by_index(data: pd.DataFrame) -> Dict[Union[str, Tuple[str, ...]], Set[int]]:
        grouped: Dict[Union[str, Tuple[str, ...]], Set[int]] = {}
        columns = data.columns.values
        for col_idx, col_name in enumerate(columns):
            if col_name[0] not in grouped:
                grouped[col_name[0]] = set()
            grouped[col_name[0]].add(col_idx)
        return grouped



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0
"""
Tabular Data Converter: Load data from multiple tables and use a mapping file to convert the data to PGM
"""
import inspect
import logging
from pathlib import Path
from typing import Any, Collection, Dict, List, Mapping, Optional, Union, cast

import numpy as np
import pandas as pd
import yaml
from power_grid_model import initialize_array
from power_grid_model.data_types import Dataset

from power_grid_model_io.converters.base_converter import BaseConverter
from power_grid_model_io.data_stores.base_data_store import BaseDataStore
from power_grid_model_io.data_types import ExtraInfo, TabularData
from power_grid_model_io.mappings.multiplier_mapping import MultiplierMapping, Multipliers
from power_grid_model_io.mappings.tabular_mapping import InstanceAttributes, Tables, TabularMapping
from power_grid_model_io.mappings.unit_mapping import UnitMapping, Units
from power_grid_model_io.mappings.value_mapping import ValueMapping, Values
from power_grid_model_io.utils.modules import get_function


class TabularConverter(BaseConverter[TabularData]):
    """Tabular Data Converter: Load data from multiple tables and use a mapping file to convert the data to PGM"""

    def __init__(
        self,
        mapping_file: Optional[Path] = None,
        source: Optional[BaseDataStore[TabularData]] = None,
        destination: Optional[BaseDataStore[TabularData]] = None,
        log_level: int = logging.INFO,
    ):
        """
        Prepare some member variables and optionally load a mapping file

        Args:
            mapping_file: A yaml file containing the mapping.
        """
        super().__init__(source=source, destination=destination, log_level=log_level)
        self._mapping: TabularMapping = TabularMapping(mapping={}, logger=self._log)
        self._units: Optional[UnitMapping] = None
        self._substitutions: Optional[ValueMapping] = None
        self._multipliers: Optional[MultiplierMapping] = None
        if mapping_file is not None:
            self.set_mapping_file(mapping_file=mapping_file)

    def set_mapping_file(self, mapping_file: Path) -> None:
        """Read, parse and interpret a mapping file.
        Args:
          mapping_file: The path to the mapping file

        """
        # Read mapping
        if mapping_file.suffix.lower() != ".yaml":
            raise ValueError(f"Mapping file should be a .yaml file, {mapping_file.suffix} provided.")
        self._log.debug("Read mapping file", mapping_file=mapping_file)

        with open(mapping_file, "r", encoding="utf-8") as mapping_stream:
            mapping = yaml.safe_load(mapping_stream)

        if "grid" not in mapping:
            raise KeyError("Missing 'grid' mapping in mapping_file")

        self.set_mapping(mapping=mapping)

    def set_mapping(self, mapping: Mapping[str, Any]) -> None:
        """Interpret a mapping structure. This includes:
         * the table to table mapping ('grid')
         * the unit conversions ('units')
         * the value substitutions ('substitutions') (e.g. enums or other one-on-one value mapping)
         * column multipliers ('multipliers') (e.g. values in a column ending with _kv should be multiplied by 1000.0)

        Args:
          mapping: A mapping structure (dictionary):

        """
        if "grid" in mapping:
            self._mapping = TabularMapping(cast(Tables, mapping["grid"]), logger=self._log)
        if "units" in mapping:
            self._units = UnitMapping(cast(Units, mapping["units"]), logger=self._log)
        if "substitutions" in mapping:
            self._substitutions = ValueMapping(cast(Values, mapping["substitutions"]), logger=self._log)
        if "multipliers" in mapping:
            self._multipliers = MultiplierMapping(cast(Multipliers, mapping["multipliers"]), logger=self._log)

    def _parse_data(self, data: TabularData, data_type: str, extra_info: Optional[ExtraInfo]) -> Dataset:
        """This function parses tabular data and returns power-grid-model data

        Args:
          data: TabularData, i.e. a dictionary with the components as keys and pd.DataFrames as values, with
        attribute names as columns and their values in the table
          data_type: power-grid-model data type, i.e. "input" or "update"
          extra_info: an optional dictionary where extra component info (that can't be specified in
        power-grid-model data) can be specified
          data: TabularData:
          data_type: str:
          extra_info: Optional[ExtraInfo]:

        Returns:
          a power-grid-model dataset, i.e. a dictionary as {component: np.ndarray}

        """
        # Apply units and substitutions to the data. Note that the conversions are 'lazy', i.e. the units and
        # substitutions will be applied the first time .get_column(table, field) is called.
        if self._units is not None:
            data.set_unit_multipliers(self._units)
        if self._substitutions is not None:
            data.set_substitutions(self._substitutions)

        # Initialize some empty data structures
        pgm: Dict[str, List[np.ndarray]] = {}

        # For each table in the mapping
        for table in self._mapping.tables():
            if table not in data or len(data[table]) == 0:
                continue  # pragma: no cover (bug in python 3.9)
            for component, attributes in self._mapping.instances(table=table):
                component_data = self._convert_table_to_component(
                    data=data,
                    data_type=data_type,
                    table=table,
                    component=component,
                    attributes=attributes,
                    extra_info=extra_info,
                )
                if component_data is not None:
                    if component not in pgm:
                        pgm[component] = []
                    pgm[component].append(component_data)

        input_data = TabularConverter._merge_pgm_data(data=pgm)
        self._log.debug(
            "Converted tabular data to power grid model data",
            n_components=len(input_data),
            n_instances=sum(len(table) for table in input_data.values()),
        )
        return input_data

    # pylint: disable = too-many-arguments
    def _convert_table_to_component(
        self,
        data: TabularData,
        data_type: str,
        table: str,
        component: str,
        attributes: InstanceAttributes,
        extra_info: Optional[ExtraInfo],
    ) -> Optional[np.ndarray]:
        """
        This function converts a single table/sheet of TabularData to a power-grid-model input/update array. One table
        corresponds to one component

        Args:
          data: The full dataset with tabular data
          data_type: The data type, i.e. "input" or "update"
          table: The name of the table that should be converter
          component: the component for which a power-grid-model array should be made
          attributes: a dictionary with a mapping from the attribute names in the table to the corresponding
        power-grid-model attribute names
          extra_info: an optional dictionary where extra component info (that can't be specified in
        power-grid-model data) can be specified
          data: TabularData:
          data_type: str:
          table: str:
          component: str:
          attributes: InstanceAttributes:
          extra_info: Optional[ExtraInfo]:

        Returns:
          returns a power-grid-model structured array for one component

        """
        if table not in data:
            return None

        if "filters" in attributes:
            table_mask = self._parse_table_filters(data=data, table=table, filtering_functions=attributes["filters"])
            if table_mask is not None and not table_mask.any():
                return None
        else:
            table_mask = None

        n_records = np.sum(table_mask) if table_mask is not None else len(data[table])

        try:
            pgm_data = initialize_array(data_type=data_type, component_type=component, shape=n_records)
        except KeyError as ex:
            raise KeyError(f"Invalid component type '{component}' or data type '{data_type}'") from ex

        if "id" not in attributes:
            raise KeyError(f"No mapping for the attribute 'id' for '{component}s'!")

        # Make sure that the "id" column is always parsed first (at least before "extra" is parsed)
        attributes_without_filter = {k: v for k, v in attributes.items() if k != "filters"}
        sorted_attributes = sorted(attributes_without_filter.items(), key=lambda x: "" if x[0] == "id" else x[0])

        for attr, col_def in sorted_attributes:
            self._convert_col_def_to_attribute(
                data=data,
                pgm_data=pgm_data,
                table=table,
                component=component,
                attr=attr,
                col_def=col_def,
                table_mask=table_mask,
                extra_info=extra_info,
            )

        return pgm_data

    def _parse_table_filters(self, data: TabularData, table: str, filtering_functions: Any) -> Optional[np.ndarray]:
        if not isinstance(data[table], pd.DataFrame):
            return None

        table_mask = np.ones(len(data[table]), dtype=bool)
        for filtering_fn in filtering_functions:
            for fn_name, kwargs in filtering_fn.items():
                fn_ptr = get_function(fn_name)
                table_mask &= data[table].apply(fn_ptr, axis=1, **kwargs).values
        return table_mask

    # pylint: disable = too-many-arguments
    def _convert_col_def_to_attribute(
        self,
        data: TabularData,
        pgm_data: np.ndarray,
        table: str,
        component: str,
        attr: str,
        col_def: Any,
        table_mask: Optional[np.ndarray],
        extra_info: Optional[ExtraInfo],
    ):
        """This function updates one of the attributes of pgm_data, based on the corresponding table/column in a tabular
        dataset

        Args:
          data: TabularData, i.e. a dictionary with the components as keys and pd.DataFrames as values, with
        attribute names as columns and their values in the table
          pgm_data: a power-grid-model input/update array for one component
          table: the table name of the particular component in the tabular dataset
          component: the corresponding component
          attr: the name of the attribute that should be updated in the power-grid-model array
          col_def: the name of the column where the attribute values can be found
          extra_info: an optional dictionary where extra component info (that can't be specified in
        power-grid-model data) can be specified
          data: TabularData:
          pgm_data: np.ndarray:
          table: str:
          component: str:
          attr: str:
          col_def: Any:
          extra_info: Optional[ExtraInfo]:

        Returns:
          the function updates pgm_data, it should not return something

        """
        # To avoid mistakes, the attributes in the mapping should exist. There is one extra attribute called
        # 'extra' in which extra information can be captured.
        if attr not in pgm_data.dtype.names and attr not in ["extra", "filters"]:
            attrs = ", ".join(pgm_data.dtype.names)
            raise KeyError(f"Could not find attribute '{attr}' for '{component}s'. (choose from: {attrs})")

        if attr == "extra":
            # Extra info must be linked to the object IDs, therefore the uuids should be known before extra info can
            # be parsed. Before this for loop, it is checked that "id" exists and it is placed at the front.
            self._handle_extra_info(
                data=data,
                table=table,
                col_def=col_def,
                uuids=pgm_data["id"],
                table_mask=table_mask,
                extra_info=extra_info,
            )
            # Extra info should not be added to the numpy arrays, so let's continue to the next attribute
            return

        attr_data = self._parse_col_def(
            data=data, table=table, table_mask=table_mask, col_def=col_def, extra_info=extra_info
        )

        if len(attr_data.columns) != 1:
            raise ValueError(f"DataFrame for {component}.{attr} should contain a single column ({attr_data.columns})")

        pgm_data[attr] = attr_data.iloc[:, 0]

    def _handle_extra_info(
        self,
        data: TabularData,
        table: str,
        col_def: Any,
        uuids: np.ndarray,
        table_mask: Optional[np.ndarray],
        extra_info: Optional[ExtraInfo],
    ) -> None:
        """This function can extract extra info from the tabular data and store it in the extra_info dict

        Args:
          data: tabularData, i.e. a dictionary with the components as keys and pd.DataFrames as values, with
        attribute names as columns and their values in the table
          table: the table name of the particular component in the tabular dataset
          col_def: the name of the column that should be stored in extra_info
          uuids: a numpy nd.array containing the uuids of the components
          extra_info: an optional dictionary where extra component info (that can't be specified in
        power-grid-model data) can be specified
          data: TabularData:
          table: str:
          col_def: Any:
          uuids: np.ndarray:
          extra_info: Optional[ExtraInfo]:

        Returns:

        """
        if extra_info is None:
            return

        extra = self._parse_col_def(
            data=data, table=table, table_mask=table_mask, col_def=col_def, extra_info=None
        ).to_dict(orient="records")
        for i, xtr in zip(uuids, extra):
            xtr = {
                k[0] if isinstance(k, tuple) else k: v
                for k, v in xtr.items()
                if not isinstance(v, float) or not np.isnan(v)
            }
            if xtr:
                if i in extra_info:
                    extra_info[i].update(xtr)
                else:
                    extra_info[i] = xtr

    @staticmethod
    def _merge_pgm_data(data: Dict[str, List[np.ndarray]]) -> Dict[str, np.ndarray]:
        """During the conversion, multiple numpy arrays can be produced for the same type of component. These arrays
        should be concatenated to form one large table.

        Args:
          data: For each component, one or more numpy structured arrays
          data: Dict[str:
          List[np.ndarray]]:

        Returns:

        """
        merged = {}
        for component_name, data_set in data.items():
            # If there is only one array, use it as is
            if len(data_set) == 1:
                merged[component_name] = data_set[0]

            # If there are multiple arrays, concatenate them
            elif len(data_set) > 1:
                # pylint: disable=unexpected-keyword-arg
                merged[component_name] = np.concatenate(data_set, dtype=data_set[0].dtype)

        return merged

    def _serialize_data(self, data: Dataset, extra_info: Optional[ExtraInfo]) -> TabularData:
        if extra_info is not None:
            raise NotImplementedError("Extra info can not (yet) be stored for tabular data")
        if isinstance(data, list):
            raise NotImplementedError("Batch data can not (yet) be stored for tabular data")
        return TabularData(logger=self._log, **data)

    def _parse_col_def(
        self,
        data: TabularData,
        table: str,
        col_def: Any,
        table_mask: Optional[np.ndarray],
        extra_info: Optional[ExtraInfo],
    ) -> pd.DataFrame:
        """Interpret the column definition and extract/convert/create the data as a pandas DataFrame.

        Args:
          data: TabularData:
          table: str:
          col_def: Any:
          extra_info: Optional[ExtraInfo]:

        Returns:

        """
        if isinstance(col_def, (int, float)):
            return self._parse_col_def_const(data=data, table=table, col_def=col_def, table_mask=table_mask)
        if isinstance(col_def, str):
            return self._parse_col_def_column_name(data=data, table=table, col_def=col_def, table_mask=table_mask)
        if isinstance(col_def, dict):
            return self._parse_col_def_filter(
                data=data, table=table, table_mask=table_mask, col_def=col_def, extra_info=extra_info
            )
        if isinstance(col_def, list):
            return self._parse_col_def_composite(data=data, table=table, col_def=col_def, table_mask=table_mask)
        raise TypeError(f"Invalid column definition: {col_def}")

    @staticmethod
    def _parse_col_def_const(
        data: TabularData, table: str, col_def: Union[int, float], table_mask: Optional[np.ndarray] = None
    ) -> pd.DataFrame:
        """Create a single column pandas DataFrame containing the const value.

        Args:
          data: TabularData:
          table: str:
          col_def: Union[int:
          float]:

        Returns:

        """
        assert isinstance(col_def, (int, float))
        const_df = pd.DataFrame([col_def] * len(data[table]))
        if table_mask is not None:
            # Required to retain indices before filter
            return const_df[table_mask]
        return const_df

    def _parse_col_def_column_name(
        self, data: TabularData, table: str, col_def: str, table_mask: Optional[np.ndarray] = None
    ) -> pd.DataFrame:
        """Extract a column from the data. If the column doesn't exist, check if the col_def is a special float value,
        like 'inf'. If that's the case, create a single column pandas DataFrame containing the const value.

        Args:
          data: TabularData:
          table: str:
          col_def: str:

        Returns:

        """
        assert isinstance(col_def, str)

        table_data = data[table]
        if table_mask is not None:
            table_data = table_data[table_mask]

        # If multiple columns are given in col_def, return the first column that exists in the dataset
        columns = [col_name.strip() for col_name in col_def.split("|")]
        for col_name in columns:
            if col_name in table_data or col_name == "index":
                col_data = data.get_column(table_name=table, column_name=col_name)
                if table_mask is not None:
                    col_data = col_data[table_mask]
                col_data = self._apply_multiplier(table=table, column=col_name, data=col_data)
                return pd.DataFrame(col_data)

        try:  # Maybe it is not a column name, but a float value like 'inf', let's try to convert the string to a float
            const_value = float(col_def)
        except ValueError:
            # pylint: disable=raise-missing-from
            columns_str = " and ".join(f"'{col_name}'" for col_name in columns)
            raise KeyError(f"Could not find column {columns_str} on table '{table}'")

        return self._parse_col_def_const(data=data, table=table, col_def=const_value, table_mask=table_mask)

    def _apply_multiplier(self, table: str, column: str, data: pd.Series) -> pd.Series:
        if self._multipliers is None:
            return data
        try:
            multiplier = self._multipliers.get_multiplier(table=table, attr=column)
            self._log.debug("Applied multiplier", table=table, column=column, multiplier=multiplier)
            return data * multiplier
        except KeyError:
            return data

    def _parse_reference(
        self,
        data: TabularData,
        table: str,
        other_table: str,
        query_column: str,
        key_column: str,
        value_column: str,
        table_mask: Optional[np.ndarray],
    ) -> pd.DataFrame:
        """
        Find and extract a column from a different table.

        Args:
            data: The data
            table: The current table named
            other_table: The table in which we would like to find a value
            key_column: The column in the current table that stores the keys
            query_column: The column in the other table in which we should look for the keys
            value_column: The column in the other table which stores the values that we would like to return

        Returns:

        """
        queries = self._parse_col_def_column_name(data=data, table=table, col_def=query_column, table_mask=table_mask)
        keys = self._parse_col_def_column_name(data=data, table=other_table, col_def=key_column, table_mask=None)
        values = self._parse_col_def_column_name(data=data, table=other_table, col_def=value_column, table_mask=None)
        other = pd.concat([keys, values], axis=1)
        result = queries.merge(other, how="left", left_on=query_column, right_on=key_column)
        return result[[value_column]]

    def _parse_col_def_filter(
        self,
        data: TabularData,
        table: str,
        col_def: Dict[str, Any],
        table_mask: Optional[np.ndarray],
        extra_info: Optional[ExtraInfo],
    ) -> pd.DataFrame:
        """
        Parse column filters like 'auto_id', 'reference', 'function', etc
        """
        assert isinstance(col_def, dict)
        data_frames = []
        for name, sub_def in col_def.items():
            if name == "auto_id":
                # Check that "key" is in the definition and no other keys than "table" and "name"
                if (
                    not isinstance(sub_def, dict)
                    or "key" not in sub_def
                    or len(set(sub_def.keys()) & {"table", "name"}) > 2
                ):
                    raise ValueError(f"Invalid {name} definition: {sub_def}")
                col_data = self._parse_auto_id(
                    data=data,
                    table=table,
                    table_mask=table_mask,
                    ref_table=sub_def.get("table"),
                    ref_name=sub_def.get("name"),
                    key_col_def=sub_def["key"],
                    extra_info=extra_info,
                )
            elif name == "reference":
                # Check that (only) the required keys are in the definition
                if not isinstance(sub_def, dict) or {
                    "other_table",
                    "query_column",
                    "key_column",
                    "value_column",
                } != set(sub_def.keys()):
                    raise ValueError(f"Invalid {name} definition: {sub_def}")
                return self._parse_reference(
                    data=data,
                    table=table,
                    table_mask=table_mask,
                    other_table=sub_def["other_table"],
                    query_column=sub_def["query_column"],
                    key_column=sub_def["key_column"],
                    value_column=sub_def["value_column"],
                )
            elif isinstance(sub_def, list):
                col_data = self._parse_pandas_function(
                    data=data, table=table, table_mask=table_mask, fn_name=name, col_def=sub_def
                )
            elif isinstance(sub_def, dict):
                col_data = self._parse_function(
                    data=data, table=table, table_mask=table_mask, function=name, col_def=sub_def
                )
            else:
                raise TypeError(f"Invalid {name} definition: {sub_def}")
            data_frames.append(col_data)
        return pd.concat(data_frames, axis=1)

    def _parse_auto_id(
        self,
        data: TabularData,
        table: str,
        ref_table: Optional[str],
        ref_name: Optional[str],
        key_col_def: Union[str, List[str], Dict[str, str]],
        table_mask: Optional[np.ndarray],
        extra_info: Optional[ExtraInfo],
    ) -> pd.DataFrame:
        """
        Create (or retrieve) a unique numerical id for each object (row) in `data[table]`, based on the `name`
        attribute, which is constant for each object, and the value(s) of `key_col_def`, which describes most likely a
        single column, or a list of columns.

        Args:
            data: The entire input data
            table: The current table name
            ref_table: The table name to which the id refers. If None, use the current table name.
            ref_name: A custom textual identifier, to be used for the auto_id. If None, ignore it.
            key_col_def: A column definition which should be unique for each object within the current table

        Returns: A single column containing numerical ids

        """

        # Handle reference table
        # mypy complains about ref_table being optional, therefore ref_table_str is defined as a string
        ref_table_str = ref_table or table

        # Handle reference column definition
        if isinstance(key_col_def, dict):
            key_names = list(key_col_def.keys())
            key_col_def = list(key_col_def.values())
        elif isinstance(key_col_def, list):
            key_names = key_col_def
        elif isinstance(key_col_def, str):
            key_names = [key_col_def]
        else:
            raise TypeError(f"Invalid key definition type '{type(key_col_def).__name__}': {key_col_def}")

        col_data = self._parse_col_def(
            data=data, table=table, table_mask=table_mask, col_def=key_col_def, extra_info=None
        )

        def auto_id(row: np.ndarray):
            key = dict(zip(key_names, row))
            pgm_id = self._get_id(table=ref_table_str, key=key, name=ref_name)

            # Extra info should only be added for the "id" field. Unfortunately we cannot check the field name at
            # this point, so we'll use a heuristic:
            # 1. An extra info dictionary should be supplied (i.e. extra info is requested by the caller).
            # 2. The auto_id should refer to the current table.
            #    (a counter example is an auto id referring to the nodes table, while the current table is lines)
            # 3. There shouldn't be any extra info for the current pgm_id, because the id attribute is supposed to be
            #    the first argument to be parsed.
            if extra_info is not None and ref_table_str == table and pgm_id not in extra_info:
                if ref_name is not None:
                    extra_info[pgm_id] = {"id_reference": {"table": ref_table_str, "name": ref_name, "key": key}}
                else:
                    extra_info[pgm_id] = {"id_reference": {"table": ref_table_str, "key": key}}

            return pgm_id

        if col_data.empty:
            return col_data
        return col_data.apply(auto_id, axis=1, raw=True)

    def _parse_pandas_function(
        self, data: TabularData, table: str, fn_name: str, col_def: List[Any], table_mask: Optional[np.ndarray]
    ) -> pd.DataFrame:
        """Special vectorized functions.

        Args:
          data: The data
          table: The name of the current table
          fn_name: The name of the function.
          col_def: The definition of the function arguments

        Returns:

        """
        assert isinstance(col_def, list)

        # "multiply" is an alias for "prod"
        if fn_name == "multiply":
            fn_name = "prod"

        col_data = self._parse_col_def(data=data, table=table, col_def=col_def, table_mask=table_mask, extra_info=None)

        try:
            fn_ptr = getattr(col_data, fn_name)
        except AttributeError as ex:
            raise ValueError(f"Pandas DataFrame has no function '{fn_name}'") from ex

        # If the function expects an 'other' argument, apply the function per column (e.g. divide)
        empty = inspect.Parameter.empty
        fn_sig = inspect.signature(fn_ptr)
        if "other" in fn_sig.parameters and fn_sig.parameters["other"].default == empty:
            n_columns = col_data.shape[1] if col_data.ndim > 1 else 1
            result = col_data.iloc[:, 0]
            for i in range(1, n_columns):
                result = getattr(result, fn_name)(other=col_data.iloc[:, i])
            return pd.DataFrame(result)

        # If the function expects any argument
        if any(param.default == empty for name, param in fn_sig.parameters.items() if name != "kwargs"):
            raise ValueError(f"Invalid pandas function DataFrame.{fn_name}")

        return pd.DataFrame(fn_ptr(axis=1))

    def _parse_function(
        self, data: TabularData, table: str, function: str, col_def: Dict[str, Any], table_mask: Optional[np.ndarray]
    ) -> pd.DataFrame:
        """Import the function by name and apply it to each row.

        Args:
          data: The data
          table: The name of the current table
          function: The name (or path) of the function.
          col_def: The definition of the function keyword arguments

        Returns:

        """
        assert isinstance(col_def, dict)

        fn_ptr = get_function(function)
        key_words = list(col_def.keys())
        sub_def = list(col_def.values())
        col_data = self._parse_col_def(data=data, table=table, col_def=sub_def, table_mask=table_mask, extra_info=None)

        if col_data.empty:
            raise ValueError(f"Cannot apply function {function} to an empty DataFrame")

        col_data = col_data.apply(lambda row, fn=fn_ptr: fn(**dict(zip(key_words, row))), axis=1, raw=True)
        return pd.DataFrame(col_data)

    def _parse_col_def_composite(
        self, data: TabularData, table: str, col_def: list, table_mask: Optional[np.ndarray]
    ) -> pd.DataFrame:
        """Select multiple columns (each is created from a column definition) and return them as a new DataFrame.

        Args:
          data: TabularData:
          table: str:
          col_def: list:

        Returns:

        """
        assert isinstance(col_def, list)
        columns = [
            self._parse_col_def(data=data, table=table, col_def=sub_def, table_mask=table_mask, extra_info=None)
            for sub_def in col_def
        ]
        return pd.concat(columns, axis=1)

    def _get_id(self, table: str, key: Mapping[str, int], name: Optional[str]) -> int:
        """
        Get a unique numerical ID for the supplied name / key combination
        Args:
            table: Table name (e.g. "Nodes")
            key: Component identifier (e.g. {"name": "node1"} or {"number": 1, "sub_number": 2})
            name: Optional component name (e.g. "internal_node")
        Returns: A unique id
        """
        auto_id_key = (table, tuple(sorted(key.items())), name)
        return self._auto_id(item=(table, key, name), key=auto_id_key)

    def get_id(self, table: str, key: Mapping[str, int], name: Optional[str] = None) -> int:
        """
        Get a the numerical ID previously associated with the supplied name / key combination
        Args:
            table: Table name (e.g. "Nodes")
            key: Component identifier (e.g. {"name": "node1"} or {"number": 1, "sub_number": 2})
            name: Optional component name (e.g. "internal_node")
        Returns: The associated id
        """
        auto_id_key = (table, tuple(sorted(key.items())), name)
        if auto_id_key not in self._auto_id:
            raise KeyError((table, key, name))
        return self._auto_id(item=(table, key, name), key=auto_id_key)

    def get_ids(self, keys: pd.DataFrame, table: Optional[str] = None, name: Optional[str] = None) -> List[int]:
        """
        Get a the numerical ID previously associated with the supplied name / key combination
        Args:
            keys: Component identifiers (e.g. a pandas Dataframe with columns "number" and "sub_number")
            table: Table name (e.g. "Nodes")
            name: Optional component name (e.g. "internal_node")
        Returns: The associated id
        """

        def get_id(row: pd.Series) -> int:
            key = row.dropna().to_dict()
            row_table = key.pop("table") if table is None and "table" in key else table
            row_name = key.pop("name") if name is None and "name" in key else name
            return self.get_id(table=row_table, key=key, name=row_name)

        return keys.apply(get_id, axis=1).to_list()

    def lookup_id(self, pgm_id: int) -> Dict[str, Union[str, Dict[str, int]]]:
        """
        Retrieve the original name / key combination of a pgm object
        Args:
            pgm_id: a unique numerical ID
        Returns: The original name / key combination
        """
        table, key, name = self._auto_id[pgm_id]
        reference = {"table": table}
        if name is not None:
            reference["name"] = name
        reference["key"] = key
        return reference

    def lookup_ids(self, pgm_ids: Collection[int]) -> pd.DataFrame:
        """
        Retrieve the original name / key combination of a list of pgm object
        Args:
            pgm_ids: a list of unique numerical ID
        Returns: A (possibly sparse) pandas dataframe storing all the original reference data
        """

        def lookup(pgm_id):
            table, key, name = self._auto_id[pgm_id]
            reference = {"table": table}
            if name is not None:
                reference["name"] = name
            reference.update(key)
            return reference

        return pd.DataFrame((lookup(pgm_id) for pgm_id in pgm_ids), index=pgm_ids)
